import copy
import logging
import numpy as np
from libensemble.resources.resources import Resources

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class InsufficientResourcesError(Exception):
    "Raised when more resources are requested than exist"


class InsufficientFreeResources(Exception):
    "Raised when the requested resources could not be provisioned"


class ResourceScheduler:
    """Calculates and returns resource set ids from a dictionary of resource sets
    by group. Resource sets available is read from the resources module or from
    a resources object passed in."""

    def __init__(self, user_resources=None, sched_opts={}):
        """Initiate a ResourceScheduler object

        Parameters
        ----------

        user_resources: Resources object, optional
            A resources object. If present overrides the class variable.

        sched_opts: dict, optional
            A dictionary of scheduler options. Normally passed via alloc_specs['user']


        **Scheduler Options**:

        split2fit: boolean, optional
            Split across more nodes if space not currently avail (even though can fit when free).
            Default: True
        """

        self.resources = user_resources or Resources.resources.resource_manager
        self.rsets_free = self.resources.rsets_free
        self.avail_rsets_by_group = None

        # Process scheduler options
        self.split2fit = sched_opts.get('split2fit', True)

    def assign_resources(self, rsets_req):
        """Schedule resource sets to a work item if possible

        If the resources required are less than one node, they will be
        allocated to the smallest available sufficient slot.

        If the resources required are more than one node, then will
        attempt to find an even split. If no even split is possible,
        then enough full nodes will be allocated to cover the requirement.

        Returns a list of resource sets ids. A return of None implies
        insufficient resources.
        """

        if rsets_req > self.resources.total_num_rsets:
            raise InsufficientResourcesError("More resource sets requested {} than exist {}"
                                             .format(rsets_req, self.resources.total_num_rsets))

        if rsets_req > self.rsets_free:
            raise InsufficientFreeResources

        num_groups = self.resources.num_groups
        max_grpsize = self.resources.rsets_per_node  # assumes even
        avail_rsets_by_group = self.get_avail_rsets_by_group()

        # Work out best target fit - if all rsets were free.
        rsets_req, num_groups_req, rsets_req_per_group = \
            self.calc_req_split(rsets_req, max_grpsize, num_groups, extend=True)

        if self.split2fit:
            sorted_lengths = ResourceScheduler.get_sorted_lens(avail_rsets_by_group)
            max_even_grpsize = sorted_lengths[num_groups_req - 1]
            if max_even_grpsize == 0 and rsets_req > 0:
                raise InsufficientFreeResources
            if max_even_grpsize < rsets_req_per_group:
                # Cannot fit in smallest number of nodes - try to split
                rsets_req, num_groups_req, rsets_req_per_group = \
                    self.calc_even_split_uneven_groups(max_even_grpsize, num_groups_req,
                                                       rsets_req, sorted_lengths, num_groups, extend=False)
        tmp_avail_rsets_by_group = copy.deepcopy(avail_rsets_by_group)
        max_upper_bound = max_grpsize + 1

        # Now find slots on as many nodes as need
        accum_team = []
        group_list = []

        for ng in range(num_groups_req):
            cand_team, cand_group = \
                self.find_candidate(tmp_avail_rsets_by_group, group_list, rsets_req_per_group, max_upper_bound)

            if cand_group is not None:
                accum_team.extend(cand_team)
                group_list.append(cand_group)

                for rset in cand_team:
                    tmp_avail_rsets_by_group[cand_group].remove(rset)

        if len(accum_team) == rsets_req:
            # A successful team found
            rset_team = sorted(accum_team)
            self.avail_rsets_by_group = tmp_avail_rsets_by_group
            self.rsets_free -= rsets_req
        else:
            raise InsufficientFreeResources

        return rset_team

    def find_candidate(self, rsets_by_group, group_list, rsets_req_per_group, max_upper_bound):
        """Find a candidate slot in a group"""
        cand_team = []
        cand_group = None
        upper_bound = max_upper_bound
        for g in rsets_by_group:
            if g in group_list:
                continue
            nslots = len(rsets_by_group[g])
            if nslots == rsets_req_per_group:
                # Exact fit
                cand_team = rsets_by_group[g].copy()
                cand_group = g
                break
            elif rsets_req_per_group < nslots < upper_bound:
                cand_team = rsets_by_group[g][:rsets_req_per_group]
                cand_group = g
                upper_bound = nslots
        return cand_team, cand_group

    def get_avail_rsets_by_group(self):
        """Return a dictionary of resource set IDs for each group (e.g. node)

        If groups are not set they will all be in one group (group 0)

        E.g: Say 8 resource sets / 2 nodes
        GROUP  1: [1,2,3,4]
        GROUP  2: [5,6,7,8]
        """
        if self.avail_rsets_by_group is None:
            rsets = self.resources.rsets
            groups = np.unique(rsets['group'])
            self.avail_rsets_by_group = {}
            for g in groups:
                self.avail_rsets_by_group[g] = []
            for ind, rset in enumerate(rsets):
                if not rset['assigned']:
                    g = rset['group']
                    self.avail_rsets_by_group[g].append(ind)
        return self.avail_rsets_by_group

    def calc_req_split(self, rsets_req, max_grpsize, num_groups, extend):
        if self.resources.even_groups:  # This is total group sizes even (not available sizes)
            rsets_req, num_groups_req, rsets_req_per_group = \
                self.calc_rsets_even_grps(rsets_req, max_grpsize, num_groups, extend)
        else:
            logger.warning('Uneven groups - but using even groups function')
            rsets_req, num_groups_req, rsets_req_per_group = \
                self.calc_rsets_even_grps(rsets_req, max_grpsize, num_groups, extend)
        return rsets_req, num_groups_req, rsets_req_per_group

    def calc_rsets_even_grps(self, rsets_req, max_grpsize, max_groups, extend):
        """Calculate an even breakdown to best fit rsets_req input"""
        if rsets_req == 0:
            return 0, 0, 0

        num_groups_req = rsets_req//max_grpsize + (rsets_req % max_grpsize > 0)  # Divide with roundup.

        # Up to max groups - keep trying for an even split
        if num_groups_req > 1:
            even_partition = False
            tmp_num_groups = num_groups_req
            while tmp_num_groups <= max_groups:
                if rsets_req % tmp_num_groups == 0:
                    even_partition = True
                    break
                tmp_num_groups += 1

            if even_partition:
                num_groups_req = tmp_num_groups
                rsets_req_per_group = rsets_req//num_groups_req  # This should always divide perfectly.
            else:
                if extend:
                    rsets_req_per_group = rsets_req//num_groups_req + (rsets_req % num_groups_req > 0)
                    rsets_req = num_groups_req * rsets_req_per_group
                    logger.debug('Increasing resource requirement to obtain an even partition of resource sets'
                                 '\nto nodes. rsets_req {}  num_groups_req {} rsets_req_per_group {}'.
                                 format(rsets_req, num_groups_req, rsets_req_per_group))
                else:
                    rsets_req_per_group = max_grpsize
        else:
            rsets_req_per_group = rsets_req
        return rsets_req, num_groups_req, rsets_req_per_group

    def calc_even_split_uneven_groups(self, rsets_per_grp, ngroups, rsets_req, sorted_lens, max_grps, extend):
        """Calculate an even breakdown to best fit rsets_req with uneven groups"""
        if rsets_req == 0:
            return 0, 0, 0

        while rsets_per_grp * ngroups != rsets_req:
            if rsets_per_grp * ngroups > rsets_req:
                rsets_per_grp -= 1
            else:
                ngroups += 1
                if ngroups > max_grps:
                    raise InsufficientFreeResources
                rsets_per_grp = sorted_lens[ngroups - 1]

        return rsets_req, ngroups, rsets_per_grp

    @staticmethod
    def get_sorted_lens(avail_rsets):
        """Get max length of a list value in a dictionary"""
        return sorted([len(v) for v in avail_rsets.values()], reverse=True)



import numpy as np
from libensemble.message_numbers import EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def finite_diff_alloc(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function will give simulation work if possible, but
    otherwise start 1 persistent generator.  If all points requested by
    the persistent generator for a given (x_ind,f_ind) pair have been returned from the
    simulation evaluation, then this information is given back to the
    persistent generator (where x_ind is in range(n) and f_ind is in range(p))

    .. seealso::
        `test_persistent_fd_param_finder.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_fd_param_finder.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}
    gen_count = support.count_persis_gens()

    if len(H) and gen_count == 0:
        # The one persistent worker is done. Exiting
        return Work, persis_info, 1

    # If wid is in persistent mode, and all of its calculated values have
    # returned, give them back to wid. Otherwise, give nothing to wid
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG):

        # What (x_ind, f_ind) pairs have all of the evaluation of all n_ind
        # values complete.
        inds_not_sent_back = ~H['given_back']
        H_tmp = H[inds_not_sent_back]

        inds_to_send = np.array([], dtype=int)
        for x_ind in range(gen_specs['user']['n']):
            for f_ind in range(gen_specs['user']['p']):
                inds = np.logical_and.reduce((H_tmp['x_ind'] == x_ind, H_tmp['f_ind'] == f_ind, H_tmp['returned']))
                if sum(inds) == gen_specs['user']['nf']:
                    inds_to_send = np.append(inds_to_send, H_tmp['sim_id'][inds])

        if len(inds_to_send):
            Work[wid] = support.gen_work(wid, gen_specs['persis_in'], inds_to_send, persis_info.get(wid),
                                         persistent=True)

    points_to_evaluate = ~H['given'] & ~H['cancel_requested']
    for wid in support.avail_worker_ids(persistent=False):
        if np.any(points_to_evaluate):
            # perform sim evaluations (if they exist in History).
            sim_ids_to_send = np.nonzero(points_to_evaluate)[0][0]  # oldest point
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, persis_info.get(wid))
            except InsufficientFreeResources:
                break
            points_to_evaluate[sim_ids_to_send] = False

        elif gen_count == 0:
            # Finally, call a persistent generator as there is nothing else to do.
            try:
                Work[wid] = support.gen_work(wid, gen_specs.get('in', []), [], persis_info.get(wid), persistent=True)
            except InsufficientFreeResources:
                break
            gen_count += 1

    return Work, persis_info, 0



import numpy as np
from libensemble.executors.executor import Executor
from libensemble.sim_funcs.surmise_test_function import borehole_true
from libensemble.message_numbers import TASK_FAILED, MAN_SIGNAL_KILL, UNSET_TAG


def subproc_borehole(H, delay):
    """This evaluates the Borehole function using a subprocess
    running compiled code.

    Note that the Executor base class submit runs a
    serial process in-place. This should work on compute nodes
    so long as there are free contexts.

    """
    with open('input', 'w') as f:
        H['thetas'][0].tofile(f)
        H['x'][0].tofile(f)

    exctr = Executor.executor
    args = 'input' + ' ' + str(delay)

    task = exctr.submit(app_name='borehole', app_args=args, stdout='out.txt', stderr='err.txt')
    calc_status = exctr.polling_loop(task, delay=0.01, poll_manager=True)

    if calc_status in [MAN_SIGNAL_KILL, TASK_FAILED]:
        f = np.inf
    else:
        f = float(task.read_stdout())
    return f, calc_status


def borehole(H, persis_info, sim_specs, libE_info):
    """
    Wraps the borehole function
    Subprocess to test receiving kill signals from manager
    """
    calc_status = UNSET_TAG  # Calc_status gets printed in libE_stats.txt
    H_o = np.zeros(H['x'].shape[0], dtype=sim_specs['out'])

    # Add a delay so subprocessed borehole takes longer
    sim_id = libE_info['H_rows'][0]
    delay = 0
    if sim_id > sim_specs['user']['init_sample_size']:
        delay = 2 + np.random.normal(scale=0.5)

    f, calc_status = subproc_borehole(H, delay)

    # Failure model (excluding observations)
    if sim_id > sim_specs['user']['num_obs']:
        if (f / borehole_true(H['x'])) > 1.25:
            f = np.inf
            calc_status = TASK_FAILED
            print('Failure of sim_id {}'.format(sim_id), flush=True)

    H_o['f'] = f
    return H_o, persis_info, calc_status



"""
A persistent generator using the uncertainty quantification capabilities in
`Tasmanian <https://tasmanian.ornl.gov/>`_.
"""

import numpy as np
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def sparse_grid_batched(H, persis_info, gen_specs, libE_info):
    """
    Implements batched construction for a Tasmanian sparse grid,
    using the loop described in Tasmanian Example 09:
    `sparse grid example <https://github.com/ORNL/TASMANIAN/blob/master/InterfacePython/example_sparse_grids_09.py>`_

    """
    U = gen_specs['user']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)
    grid = U['tasmanian_init']()  # initialize the grid
    allowed_refinements = ['setAnisotropicRefinement', 'setSurplusRefinement', 'none']
    assert 'refinement' in U and U['refinement'] in allowed_refinements, \
        "Must provide a gen_specs['user']['refinement'] in: {}".format(allowed_refinements)

    while grid.getNumNeeded() > 0:
        aPoints = grid.getNeededPoints()

        H0 = np.zeros(len(aPoints), dtype=gen_specs['out'])
        H0['x'] = aPoints

        # Receive values from manager
        tag, Work, calc_in = ps.send_recv(H0)
        if tag in [STOP_TAG, PERSIS_STOP]:
            break
        aModelValues = calc_in['f']

        # Update surrogate on grid
        t = aModelValues.reshape((aModelValues.shape[0], grid.getNumOutputs()))
        t = t.flatten()
        t = np.atleast_2d(t).T
        grid.loadNeededPoints(t)

        if 'tasmanian_checkpoint_file' in U:
            grid.write(U['tasmanian_checkpoint_file'])

        # set refinement, using user['refinement'] to pick the refinement strategy
        if U['refinement'] == 'setAnisotropicRefinement':
            assert 'sType' in U
            assert 'iMinGrowth' in U
            assert 'iOutput' in U
            grid.setAnisotropicRefinement(U['sType'], U['iMinGrowth'], U['iOutput'])
        elif U['refinement'] == 'setSurplusRefinement':
            assert 'fTolerance' in U
            assert 'iOutput' in U
            assert 'sCriteria' in U
            grid.setSurplusRefinement(U['fTolerance'], U['iOutput'], U['sCriteria'])

    return H0, persis_info, FINISHED_PERSISTENT_GEN_TAG



#!/usr/bin/env python

"""
Script to clean database and create top level app/job for Balsam tests

This creates only the parent job, not the sim/gen (they are created
in libEnsemble).

Run this before each run to ensure database is in correct state

Usage: ./create_balsam_job.py <script_name>
eg: ./create_balsam_job.py test_jobexecutor.py

This is equivalent to creating an app and job using <balsam app> and <balsam job> on the commdand line

To check create app and job do:
balsam ls apps
balsam ls jobs
"""

import os
import sys

import balsam.launcher.dag as dag
from balsam.service import models


def del_jobs():
    """ Deletes all jobs """
    Job = models.BalsamJob
    deletion_objs = Job.objects.all()
    deletion_objs.delete()


def add_app(name, exepath, desc):
    """ Add application to database """
    AppDef = models.ApplicationDefinition
    app = AppDef()
    app.name = name
    app.executable = exepath
    app.description = desc
    # app.default_preprocess = '' # optional
    # app.default_postprocess = '' # optional
    app.save()


# import balsam.launcher.dag as dag
stage_in = os.getcwd()

default_script = 'test_jobexecutor.py'

# if script provided use that - else default
if len(sys.argv) > 1:
    script = sys.argv[1]
else:
    script = default_script

# print("script is", script)

script_basename = os.path.splitext(script)[0]  # rm .py extension
app_name = script_basename + '.app'

# Add app if its not already there
AppDef = models.ApplicationDefinition
app_exists = AppDef.objects.filter(name__contains=app_name)
if not app_exists:
    app_path = sys.executable + ' ' + script
    app_desc = 'Test ' + script
    add_app(app_name, app_path, app_desc)

# Delete existing jobs
del_jobs()

# Add the job
job = dag.add_job(name='job_' + script_basename,
                  workflow="libe_workflow",  # add arg for this
                  application=app_name,
                  # application_args=job.app_args,
                  num_nodes=1,
                  procs_per_node=1,
                  stage_in_url="local:/" + stage_in,
                  stage_out_url="local:/" + stage_in,  # same as in
                  stage_out_files="*.out")



import numpy as np

const = 1500


def EvaluateFunction(x, component=np.nan):
    """
    Evaluates the chained Rosenbrock function
    """

    assert not np.isnan(component), "Must give a component"

    i = component
    x1 = x[i]
    x2 = x[i+1]
    f = 100 * (x1**2 - x2)**2 + (x1-1)**2

    return f


def EvaluateJacobian(x, component=np.nan):
    """
    Evaluates the chained Rosenbrock Jacobian
    """

    df = np.zeros(len(x), dtype=float)

    assert not np.isnan(component), "Must give a component"

    i = component
    x1 = x[i]
    x2 = x[i+1]

    df[i] = 400 * x1 * (x1**2 - x2) + 2 * (x1 - 1)
    df[i+1] = -200 * (x1**2 - x2)

    return 1.0/const * df


def alt_rosenbrock_eval(H, persis_info, sim_specs, _):

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        obj_component = H['obj_component'][i]  # which f_i
        if H[i]['get_grad']:
            H_o['gradf_i'][i] = EvaluateJacobian(x, obj_component)
        else:
            H_o['f_i'][i] = EvaluateFunction(x, obj_component)

    return H_o, persis_info



"""
Runs libEnsemble with Surmise calibration test.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_surmise_calib.py
   python3 test_persistent_surmise_calib.py --nworkers 3 --comms local
   python3 test_persistent_surmise_calib.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.

This test uses the Surmise package to perform a Borehole Calibration with
selective simulation cancellation. Initial observations are modeled using
a theta at the center of a unit hypercube. The initial function values for
these are run first. As the model is updated, the generator selects previously
issued evaluations to cancel.

See more information, see tutorial:
"Borehole Calibration with Selective Simulation Cancellation"
in the libEnsemble documentation.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4
# TESTSUITE_EXTRA: true

# Requires:
#   Install Surmise package

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.gen_funcs.persistent_surmise_calib import surmise_calib as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.sim_funcs.surmise_test_function import borehole as sim_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

# from libensemble import logger
# logger.set_level('DEBUG')  # To get debug logging in ensemble.log

if __name__ == '__main__':

    nworkers, is_manager, libE_specs, _ = parse_args()

    n_init_thetas = 15  # Initial batch of thetas
    n_x = 25  # No. of x values
    nparams = 4  # No. of theta params
    ndims = 3  # No. of x coordinates.
    max_add_thetas = 50  # Max no. of thetas added for evaluation
    step_add_theta = 10  # No. of thetas to generate per step, before emulator is rebuilt
    n_explore_theta = 200  # No. of thetas to explore while selecting the next theta
    obsvar = 10 ** (-1)  # Constant for generating noise in obs

    # Batch mode until after init_sample_size (add one theta to batch for observations)
    init_sample_size = (n_init_thetas + 1) * n_x

    # Stop after max_emul_runs runs of the emulator
    max_evals = init_sample_size + max_add_thetas * n_x

    sim_specs = {
        'sim_f': sim_f,
        'in': ['x', 'thetas'],
        'out': [('f', float)],
        'user': {'num_obs': n_x},
    }

    gen_out = [
        ('x', float, ndims),
        ('thetas', float, nparams),
        ('priority', int),
        ('obs', float, n_x),
        ('obsvar', float, n_x),
    ]

    gen_specs = {
        'gen_f': gen_f,
        'persis_in': [o[0] for o in gen_out] + ['f', 'returned', 'sim_id'],
        'out': gen_out,
        'user': {
            'n_init_thetas': n_init_thetas,  # Num thetas in initial batch
            'num_x_vals': n_x,  # Num x points to create
            'step_add_theta': step_add_theta,  # No. of thetas to generate per step
            'n_explore_theta': n_explore_theta,  # No. of thetas to explore each step
            'obsvar': obsvar,  # Variance for generating noise in obs
            'init_sample_size': init_sample_size,  # Initial batch size inc. observations
            'priorloc': 1,  # Prior location in the unit cube
            'priorscale': 0.5,  # Standard deviation of prior
        },
    }

    alloc_specs = {
        'alloc_f': alloc_f,
        'user': {
            'init_sample_size': init_sample_size,
            'async_return': True,  # True = Return results to gen as they come in (after sample)
            'active_recv_gen': True,  # Persistent gen can handle irregular communications
        },
    }

    persis_info = add_unique_random_streams({}, nworkers + 1)

    # Currently just allow gen to exit if mse goes below threshold value
    # exit_criteria = {'sim_max': max_evals, 'stop_val': ('mse', mse_exit)}

    exit_criteria = {'sim_max': max_evals}  # Now just a set number of sims.

    # Perform the run
    H, persis_info, flag = libE(
        sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs=alloc_specs, libE_specs=libE_specs
    )

    if is_manager:
        print('Cancelled sims', H['sim_id'][H['cancel_requested']])
        sims_done = np.count_nonzero(H['returned'])
        save_libE_output(H, persis_info, __file__, nworkers)
        assert sims_done == max_evals, 'Num of completed simulations should be {}. Is {}'.format(max_evals, sims_done)



import numpy as np
import libensemble.gen_funcs.persistent_aposmm as al
import libensemble.tests.unit_tests.setup as setup

libE_info = {'comm': {}}


def test_persis_aposmm_localopt_test():
    _, _, gen_specs_0, _, _ = setup.hist_setup1()

    H = np.zeros(4, dtype=[('f', float), ('sim_id', bool), ('dist_to_unit_bounds', float), ('returned', bool)])
    H['returned'] = True
    H['sim_id'] = range(len(H))
    gen_specs_0['user']['localopt_method'] = 'BADNAME'
    gen_specs_0['user']['ub'] = np.ones(2)
    gen_specs_0['user']['lb'] = np.zeros(2)

    try:
        al.aposmm(H, {}, gen_specs_0, libE_info)
    except NotImplementedError:
        assert 1, "Failed because method is unknown."
    else:
        assert 0


def test_update_history_optimal():
    hist, _, _, _, _ = setup.hist_setup1(n=2)

    H = hist.H

    H['returned'] = True
    H['sim_id'] = range(len(H))
    H['f'][0] = -1e-8
    H['x_on_cube'][-1] = 1e-10

    # Perturb x_opt point to test the case where the reported minimum isn't
    # exactly in H. Also, a point in the neighborhood of x_opt has a better
    # function value.
    opt_ind = al.update_history_optimal(H['x_on_cube'][-1]+1e-12, 1, H, np.arange(len(H)))

    assert opt_ind == 9, "Wrong point declared minimum"


def test_standalone_persistent_aposmm():

    from libensemble.tests.regression_tests.support import six_hump_camel_minima as minima
    from libensemble.sim_funcs.six_hump_camel import six_hump_camel_func, six_hump_camel_grad
    from math import gamma, pi, sqrt
    from libensemble.message_numbers import FINISHED_PERSISTENT_GEN_TAG

    persis_info = {'rand_stream': np.random.RandomState(1), 'nworkers': 4}

    n = 2
    eval_max = 2000

    gen_out = [('x', float, n), ('x_on_cube', float, n), ('sim_id', int),
               ('local_min', bool), ('local_pt', bool)]

    gen_specs = {'in': ['x', 'f', 'grad', 'local_pt', 'sim_id', 'returned', 'x_on_cube', 'local_min'],
                 'out': gen_out,
                 'user': {'initial_sample_size': 100,
                          # 'localopt_method': 'LD_MMA', # Needs gradients
                          'sample_points': np.round(minima, 1),
                          'localopt_method': 'LN_BOBYQA',
                          'standalone': {'eval_max': eval_max,
                                         'obj_func': six_hump_camel_func,
                                         'grad_func': six_hump_camel_grad,
                                         },
                          'rk_const': 0.5*((gamma(1+(n/2))*5)**(1/n))/sqrt(pi),
                          'xtol_abs': 1e-6,
                          'ftol_abs': 1e-6,
                          'dist_to_bound_multiple': 0.5,
                          'max_active_runs': 6,
                          'lb': np.array([-3, -2]),
                          'ub': np.array([3, 2])}
                 }
    H = []
    H, persis_info, exit_code = al.aposmm(H, persis_info, gen_specs, libE_info)
    assert exit_code == FINISHED_PERSISTENT_GEN_TAG, "Standalone persistent_aposmm didn't exit correctly"
    assert np.sum(H['returned']) >= eval_max, "Standalone persistent_aposmm, didn't evaluate enough points"
    assert persis_info.get('run_order'), "Standalone persistent_aposmm didn't do any localopt runs"

    tol = 1e-3
    min_found = 0
    for m in minima:
        # The minima are known on this test problem.
        # We use their values to test APOSMM has identified all minima
        print(np.min(np.sum((H[H['local_min']]['x'] - m)**2, 1)), flush=True)
        if np.min(np.sum((H[H['local_min']]['x'] - m)**2, 1)) < tol:
            min_found += 1
    assert min_found >= 6, "Found {} minima".format(min_found)

    def combined_func(x):
        return six_hump_camel_func(x), six_hump_camel_grad(x)

    gen_specs['user']['standalone'].pop('obj_func')
    gen_specs['user']['standalone'].pop('grad_func')
    gen_specs['user']['standalone']['obj_and_grad_func'] = combined_func

    H = []
    persis_info = {'rand_stream': np.random.RandomState(1), 'nworkers': 3}
    H, persis_info, exit_code = al.aposmm(H, persis_info, gen_specs, libE_info)

    assert exit_code == FINISHED_PERSISTENT_GEN_TAG, "Standalone persistent_aposmm didn't exit correctly"
    assert np.sum(H['returned']) >= eval_max, "Standalone persistent_aposmm, didn't evaluate enough points"
    assert persis_info.get('run_order'), "Standalone persistent_aposmm didn't do any localopt runs"


if __name__ == "__main__":
    test_persis_aposmm_localopt_test()
    test_update_history_optimal()
    test_standalone_persistent_aposmm()



"""
Tests the persistent_independent_optimize generator function.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_independent.py
   python3 test_persistent_independent.py --nworkers 3 --comms local

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 6
# TESTSUITE_OS_SKIP: OSX

import sys
import numpy as np
import scipy.sparse as spp

from libensemble.libE import libE
from libensemble.sim_funcs.rosenbrock import rosenbrock_eval as sim_f
from libensemble.gen_funcs.persistent_independent_optimize import independent_optimize as gen_f
from libensemble.alloc_funcs.start_persistent_consensus import start_consensus_persistent_gens as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.tools.consensus_subroutines import get_k_reach_chain_matrix

nworkers, is_manager, libE_specs, _ = parse_args()
if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")
if nworkers < 5:
    sys.exit(
        'This tests requires at least 5 workers (6 MPI processes). You can \
             decrease the number of workers by modifying the number of gens and \
             communication graph @A in the calling script.'
    )

m = 16
n = 32
num_gens = 4
eps = 1e-2

# Even though we do not use consensus matrix, we still need to pass into alloc
A = spp.diags([1, 2, 2, 1]) - get_k_reach_chain_matrix(num_gens, 1)

sim_specs = {
    'sim_f': sim_f,
    'in': ['x', 'obj_component', 'get_grad'],
    'out': [
        ('f_i', float),
        ('gradf_i', float, (n,)),
    ],
}

# lb tries to avoid x[1]=-x[2], which results in division by zero in chwirut.
gen_specs = {
    'gen_f': gen_f,
    'out': [
        ('x', float, (n,)),
        ('f_i', float),
        ('eval_pt', bool),  # eval point
        ('consensus_pt', bool),  # does not require a sim
        ('obj_component', int),  # which {f_i} to eval
        ('get_grad', bool),
    ],
    'user': {
        'lb': np.array([-1.2, 1] * (n // 2)),
        'ub': np.array([-1.2, 1] * (n // 2)),
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'user': {
        'm': m,
        'num_gens': num_gens,
    },
}

persis_info = {}
persis_info = add_unique_random_streams(persis_info, nworkers + 1)
persis_info['gen_params'] = {'eps': eps}
persis_info['sim_params'] = {'const': 1}
persis_info['A'] = A

assert n == 2 * m, "@n must be double of @m"

# Perform the run
libE_specs['safe_mode'] = False

# i==0 is full run, i==1 is early termination
for i in range(2):
    if i == 0:
        exit_criteria = {'elapsed_wallclock_time': 600, 'sim_max': 1000000}
        if is_manager:
            print('=== Testing full independent optimize ===', flush=True)
    else:
        exit_criteria = {'elapsed_wallclock_time': 600, 'sim_max': 10}
        if is_manager:
            print('=== Testing independent optimize w/ stoppage ===', flush=True)

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        print('=== End algorithm ===', flush=True)

        # check we completed
        assert flag == 0

    if is_manager and i == 0:
        # compile sum of {f_i} and {x}, and check their values are bounded by O(eps)
        eval_H = H[H['eval_pt']]

        gen_ids = np.unique(eval_H['gen_worker'])
        assert len(gen_ids) == num_gens, 'Gen did not submit any function eval requests'

        F = 0
        fstar = 0

        for i, gen_id in enumerate(gen_ids):
            last_eval_idx = np.where(eval_H['gen_worker'] == gen_id)[0][-1]
            f_i = eval_H[last_eval_idx]['f_i']
            F += f_i

        assert F - fstar < eps, 'Error of {:.4e}, expected {:.4e} (assuming f*={:.4e})'.format(F - fstar, eps, fstar)



#!/usr/bin/env python
import os
import requests
import datetime

from libensemble import Ensemble
from libensemble.executors.mpi_executor import MPIExecutor

# Import DeepDriveMD components for registering with libEnsemble's executor
from deepdrivemd.sim.openmm import run_openmm
from deepdrivemd.aggregation.basic import aggregate
from deepdrivemd.models.keras_cvae import train
from deepdrivemd.selection.latest import select_model
from deepdrivemd.agents.lof import lof


def download_data(url, file):
    if file not in os.listdir('.'):
        print('Downloading ' + file + ' ...')
        out = requests.get(url)
        with open(file, 'wb') as f:
            f.write(out.content)


folded_url = 'https://raw.githubusercontent.com/DeepDriveMD/DeepDriveMD-pipeline/main/data/bba/1FME-folded.pdb'
unfolded_url = folded_url.replace('1FME-folded.pdb', 'system/1FME-unfolded.pdb')

for url in [folded_url, unfolded_url]:
    download_data(url, url.split('/')[-1])

ddmd_apps = {'molecular_dynamics': run_openmm.__file__,
             'aggregation': aggregate.__file__,
             'machine_learning': train.__file__,
             'model_selection': select_model.__file__,
             'agent': lof.__file__}

# Submit each component via a MPI runner. This Executor can be swapped with the Balsam
#  executor to potentially submit specific components to separate systems.
exctr = MPIExecutor()
ddmd = Ensemble()

ddmd.from_yaml('libE_ddmd.yaml')

ddmd.sim_specs['user']['reference_pdb_file'] = os.path.abspath(folded_url.split('/')[-1])
ddmd.gen_specs['user']['n_traj_frames'] = int(ddmd.sim_specs['user']['sim_length_ns']*1000)

for app in ddmd_apps:
    exctr.register_app(full_path=ddmd_apps[app], app_name=app)
    ddmd.gen_specs['user'][app + '_config'] = app + '.yaml'
    ddmd.gen_specs['out'].append((app + '_cstat', int))

ddmd.libE_specs['sim_dir_symlink_files'] = [os.path.abspath(unfolded_url.split('/')[-1])]
ddmd.libE_specs['ensemble_dir_path'] = \
    os.path.abspath('./ensemble_' + str(datetime.datetime.today()).replace(' ', '_').split('.')[0])

ddmd.persis_info.add_unique_streams()

ddmd.run()



from .tools import save_libE_output, add_unique_random_streams, eprint
from .check_inputs import check_inputs
from .forkable_pdb import ForkablePdb
from .parse_args import parse_args

__all__ = ['parse_args',
           'check_inputs',
           'add_unique_random_streams',
           'save_libE_output',
           'eprint',
           'ForkablePdb'
           ]



"""
libEnsemble worker class
====================================================
"""

import socket
import logging
import logging.handlers
from itertools import count
from traceback import format_exc
from traceback import format_exception_only as format_exc_msg

import numpy as np

from libensemble.message_numbers import \
    EVAL_SIM_TAG, EVAL_GEN_TAG, \
    UNSET_TAG, STOP_TAG, PERSIS_STOP, CALC_EXCEPTION
from libensemble.message_numbers import MAN_SIGNAL_FINISH, MAN_SIGNAL_KILL
from libensemble.message_numbers import calc_type_strings, calc_status_strings
from libensemble.output_directory import EnsembleDirectory

from libensemble.utils.timer import Timer
from libensemble.executors.executor import Executor
from libensemble.resources.resources import Resources
from libensemble.comms.logs import worker_logging_config
from libensemble.comms.logs import LogConfig
import cProfile
import pstats

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)
task_timing = False


def worker_main(comm, sim_specs, gen_specs, libE_specs, workerID=None, log_comm=True):
    """Evaluates calculations given to it by the manager.

    Creates a worker object, receives work from manager, runs worker,
    and communicates results. This routine also creates and writes to
    the workers summary file.

    Parameters
    ----------
    comm: communicator
        Comm object for manager communications

    sim_specs: dict
        Parameters/information for simulation calculations

    gen_specs: dict
        Parameters/information for generation calculations

    libE_specs: dict
        Parameters/information for libE operations

    workerID: int
        Manager assigned worker ID (if None, default is comm.rank)

    log_comm: boolean
        Whether to send logging over comm
    """

    if libE_specs.get('profile'):
        pr = cProfile.Profile()
        pr.enable()

    # Receive dtypes from manager
    _, dtypes = comm.recv()
    workerID = workerID or comm.rank

    # Initialize logging on comms
    if log_comm:
        worker_logging_config(comm, workerID)

    # Set up and run worker
    worker = Worker(comm, dtypes, workerID, sim_specs, gen_specs, libE_specs)
    worker.run()

    if libE_specs.get('profile'):
        pr.disable()
        profile_state_fname = 'worker_%d.prof' % (workerID)

        with open(profile_state_fname, 'w') as f:
            ps = pstats.Stats(pr, stream=f).sort_stats('cumulative')
            ps.print_stats()


######################################################################
# Worker Class
######################################################################


class WorkerErrMsg:

    def __init__(self, msg, exc):
        self.msg = msg
        self.exc = exc


class Worker:

    """The worker class provides methods for controlling sim and gen funcs

    **Object Attributes:**

    These are public object attributes.

    :ivar comm communicator:
        Comm object for manager communications

    :ivar dict dtypes:
        Dictionary containing type information for sim and gen inputs

    :ivar int workerID:
        The libensemble Worker ID

    :ivar dict sim_specs:
        Parameters/information for simulation calculations

    :ivar dict calc_iter:
        Dictionary containing counts for each type of calc (e.g. sim or gen)
    """

    def __init__(self, comm, dtypes, workerID, sim_specs, gen_specs, libE_specs):
        """Initializes new worker object

        """
        self.comm = comm
        self.dtypes = dtypes
        self.workerID = workerID
        self.sim_specs = sim_specs
        self.libE_specs = libE_specs
        self.calc_iter = {EVAL_SIM_TAG: 0, EVAL_GEN_TAG: 0}
        self._run_calc = Worker._make_runners(sim_specs, gen_specs)
        # self._calc_id_counter = count()
        Worker._set_executor(self.workerID, self.comm)
        Worker._set_resources(self.workerID, self.comm)
        self.EnsembleDirectory = EnsembleDirectory(libE_specs=libE_specs)

    @staticmethod
    def _make_runners(sim_specs, gen_specs):
        """Creates functions to run a sim or gen"""

        sim_f = sim_specs['sim_f']

        def run_sim(calc_in, persis_info, libE_info):
            """Calls the sim func."""
            return sim_f(calc_in, persis_info, sim_specs, libE_info)

        if gen_specs:
            gen_f = gen_specs['gen_f']

            def run_gen(calc_in, persis_info, libE_info):
                """Calls the gen func."""
                return gen_f(calc_in, persis_info, gen_specs, libE_info)
        else:
            run_gen = []

        return {EVAL_SIM_TAG: run_sim, EVAL_GEN_TAG: run_gen}

    @staticmethod
    def _set_rset_team(rset_team):
        """Pass new rset_team to worker resources"""
        resources = Resources.resources
        if isinstance(resources, Resources):
            resources.worker_resources.set_rset_team(rset_team)
            return True
        else:
            return False

    @staticmethod
    def _set_executor(workerID, comm):
        """Sets worker ID in the executor, return True if set"""
        exctr = Executor.executor
        if isinstance(exctr, Executor):
            exctr.set_worker_info(comm, workerID)  # When merge update
            return True
        else:
            logger.debug("No xecutor set on worker {}".format(workerID))
            return False

    @staticmethod
    def _set_resources(workerID, comm):
        """Sets worker ID in the resources, return True if set"""
        resources = Resources.resources
        if isinstance(resources, Resources):
            resources.set_worker_resources(comm.get_num_workers(), workerID)
            return True
        else:
            logger.debug("No resources set on worker {}".format(workerID))
            return False

    def _handle_calc(self, Work, calc_in):
        """Runs a calculation on this worker object.

        This routine calls the user calculations. Exceptions are caught,
        dumped to the summary file, and raised.

        Parameters
        ----------

        Work: :obj:`dict`
            :ref:`(example)<datastruct-work-dict>`

        calc_in: obj: numpy structured array
            Rows from the :ref:`history array<datastruct-history-array>`
            for processing
        """
        calc_type = Work['tag']
        self.calc_iter[calc_type] += 1

        # calc_stats stores timing and summary info for this Calc (sim or gen)
        # calc_id = next(self._calc_id_counter)

        # from output_directory.py
        if calc_type == EVAL_SIM_TAG:
            enum_desc = 'sim_id'
            calc_id = EnsembleDirectory.extract_H_ranges(Work)
        else:
            enum_desc = 'Gen no'
            # Use global gen count if available
            if Work['libE_info'].get('gen_count'):
                calc_id = str(Work['libE_info']['gen_count'])
            else:
                calc_id = str(self.calc_iter[calc_type])
        # Add a right adjust (minimum width).
        calc_id = calc_id.rjust(5, ' ')

        timer = Timer()

        try:
            logger.debug("Starting {}: {}".format(enum_desc, calc_id))
            calc = self._run_calc[calc_type]
            with timer:
                if self.EnsembleDirectory.use_calc_dirs(calc_type):
                    loc_stack, calc_dir = self.EnsembleDirectory.prep_calc_dir(Work, self.calc_iter,
                                                                               self.workerID, calc_type)
                    with loc_stack.loc(calc_dir):  # Changes to calculation directory
                        out = calc(calc_in, Work['persis_info'], Work['libE_info'])
                else:
                    out = calc(calc_in, Work['persis_info'], Work['libE_info'])

                logger.debug("Returned from user function for {} {}".format(enum_desc, calc_id))

            assert isinstance(out, tuple), \
                "Calculation output must be a tuple."
            assert len(out) >= 2, \
                "Calculation output must be at least two elements."

            calc_status = out[2] if len(out) >= 3 else UNSET_TAG

            if calc_status is None:
                calc_status = UNSET_TAG

            # Check for buffered receive
            if self.comm.recv_buffer:
                tag, message = self.comm.recv()
                if tag in [STOP_TAG, PERSIS_STOP]:
                    if message is MAN_SIGNAL_FINISH:
                        calc_status = MAN_SIGNAL_FINISH

            return out[0], out[1], calc_status
        except Exception as e:
            logger.debug("Re-raising exception from calc {}".format(e))
            calc_status = CALC_EXCEPTION
            raise
        finally:
            # This was meant to be handled by calc_stats module.
            if task_timing and Executor.executor.list_of_tasks:
                # Initially supporting one per calc. One line output.
                task = Executor.executor.list_of_tasks[-1]
                calc_msg = "{} {}: {} {} {} Status: {}".\
                    format(enum_desc,
                           calc_id,
                           calc_type_strings[calc_type],
                           timer,
                           task.timer,
                           calc_status_strings.get(calc_status, "Not set"))
            else:
                calc_msg = "{} {}: {} {} Status: {}".\
                    format(enum_desc,
                           calc_id,
                           calc_type_strings[calc_type],
                           timer,
                           calc_status_strings.get(calc_status, "Not set"))

            logging.getLogger(LogConfig.config.stats_name).info(calc_msg)

    def _recv_H_rows(self, Work):
        """Unpacks Work request and receives any history rows"""

        libE_info = Work['libE_info']
        calc_type = Work['tag']
        if len(libE_info['H_rows']) > 0:
            _, calc_in = self.comm.recv()
        else:
            calc_in = np.zeros(0, dtype=self.dtypes[calc_type])

        logger.debug("Received calc_in ({}) of len {}".
                     format(calc_type_strings[calc_type], np.size(calc_in)))
        assert calc_type in [EVAL_SIM_TAG, EVAL_GEN_TAG], \
            "calc_type must either be EVAL_SIM_TAG or EVAL_GEN_TAG"

        return libE_info, calc_type, calc_in

    def _handle(self, Work):
        """Handles a work request from the manager"""

        # Check work request and receive second message (if needed)
        libE_info, calc_type, calc_in = self._recv_H_rows(Work)

        # Call user function
        libE_info['comm'] = self.comm
        libE_info['workerID'] = self.workerID
        libE_info['rset_team'] = libE_info.get('rset_team', [])
        Worker._set_rset_team(libE_info['rset_team'])

        calc_out, persis_info, calc_status = self._handle_calc(Work, calc_in)

        if 'libE_info' in Work:
            libE_info = Work['libE_info']

        if 'comm' in libE_info:
            del libE_info['comm']

        # If there was a finish signal, bail
        if calc_status == MAN_SIGNAL_FINISH:
            return None

        # Otherwise, send a calc result back to manager
        logger.debug("Sending to Manager with status {}".format(calc_status))
        return {'calc_out': calc_out,
                'persis_info': persis_info,
                'libE_info': libE_info,
                'calc_status': calc_status,
                'calc_type': calc_type}

    def run(self):
        """Runs the main worker loop."""

        try:
            logger.info("Worker {} initiated on node {}".
                        format(self.workerID, socket.gethostname()))

            for worker_iter in count(start=1):
                logger.debug("Iteration {}".format(worker_iter))

                mtag, Work = self.comm.recv()

                if mtag == STOP_TAG:
                    if Work is MAN_SIGNAL_FINISH:
                        break
                    elif Work is MAN_SIGNAL_KILL:
                        continue

                # Active recv is for persistent worker only - throw away here
                if Work.get('libE_info', False):
                    if Work['libE_info'].get('active_recv', False) and not Work['libE_info'].get('persistent', False):
                        if len(Work['libE_info']['H_rows']) > 0:
                            _, _, _ = self._recv_H_rows(Work)
                        continue

                response = self._handle(Work)
                if response is None:
                    break
                self.comm.send(0, response)

        except Exception as e:
            self.comm.send(0, WorkerErrMsg(' '.join(format_exc_msg(type(e), e)).strip(), format_exc()))
        else:
            self.comm.kill_pending()
        finally:
            self.EnsembleDirectory.copy_back()



import numpy as np


def gen_random_sample(H, persis_info, gen_specs, _):
    # underscore parameter for internal/testing arguments

    # Pull out user parameters to perform calculations
    user_specs = gen_specs['user']

    # Get lower and upper bounds from gen_specs
    lower = user_specs['lower']
    upper = user_specs['upper']

    # Determine how many values to generate
    num = len(lower)
    batch_size = user_specs['gen_batch_size']

    # Create array of 'batch_size' zeros
    out = np.zeros(batch_size, dtype=gen_specs['out'])

    # Replace those zeros with the random numbers
    out['x'] = persis_info['rand_stream'].uniform(lower, upper, (batch_size, num))

    # Send back our output and persis_info
    return out, persis_info



import libensemble.tests.unit_tests.setup as setup
from libensemble.message_numbers import WORKER_DONE
import numpy as np
from numpy import inf

if tuple(np.__version__.split('.')) >= ('1', '15'):
    from numpy.lib.recfunctions import repack_fields

# Consider fixtures for this - parameterization may save duplication if always use pytest.

# Comparing hist produced: options (using mix of first two)
# - hardcode array compare
# - compare selected values
# - compare from npy file - stored

wrs_H0 = np.array([(False, 0., 0, 0., 1, True, 1, True, [0., 0., 0.], True, 0.1, 1.1, False, False, False, inf),
                   (False, 0., 0, 0., 1, True, 2, True, [0., 0., 0.], True, 0.2, 1.2, False, False, False, inf),
                   (False, 0., 0, 0., 1, True, 3, True, [0., 0., 0.], True, 0.3, 1.3, False, False, False, inf)],
                  dtype=[('local_pt', '?'), ('priority', '<f8'), ('gen_worker', '<i8'), ('x_on_cube', '<f8'),
                         ('sim_id', '<i8'), ('given', '?'), ('sim_worker', '<i8'), ('returned', '?'),
                         ('fvec', '<f8', (3,)), ('allocated', '?'), ('f', '<f8'), ('given_time', '<f8'),
                         ('cancel_requested', '?'), ('kill_sent', '?'), ('given_back', '?'), ('last_given_back_time', '<f8')])

exp_H0_H = np.array([(False, 0., 0, 0., 1, True, 1, True, [0., 0., 0.], True, 0.1, 1.1, 2.1, False, False, False, inf),
                     (False, 0., 0, 0., 1, True, 2, True, [0., 0., 0.], True, 0.2, 1.2, 2.3, False, False, False, inf),
                     (False, 0., 0, 0., 1, True, 3, True, [0., 0., 0.], True, 0.3, 1.3, 2.3, False, False, False, inf),
                     (False, 0., 0, 0., -1, False, 0, False, [0., 0., 0.], False, 0., inf, 0., False, False, False, inf),
                     (False, 0., 0, 0., -1, False, 0, False, [0., 0., 0.], False, 0., inf, 0., False, False, False, inf)],
                    dtype=[('local_pt', '?'), ('priority', '<f8'), ('gen_worker', '<i8'), ('x_on_cube', '<f8'),
                           ('sim_id', '<i8'), ('given', '?'), ('sim_worker', '<i8'), ('returned', '?'),
                           ('fvec', '<f8', (3,)), ('allocated', '?'), ('f', '<f8'), ('given_time', '<f8'),
                           ('returned_time', '<f8'), ('cancel_requested', '?'), ('kill_sent', '?'),
                           ('given_back', '?'), ('last_given_back_time', '<f8')])

wrs = np.array([(False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf),
                (False, 0., 0, 0., 0., 0., -1, False, 0, False, 0, [0., 0., 0.], False, 0., inf, inf, False, 0, False, False, False, inf)],
               dtype=[('local_pt', '?'), ('priority', '<f8'), ('gen_worker', '<i8'), ('gen_time', '<f8'),
                      ('last_gen_time', '<f8'), ('x_on_cube', '<f8'), ('sim_id', '<i8'), ('given', '?'),
                      ('sim_worker', '<i8'), ('returned', '?'), ('returned_time', '<f8'), ('fvec', '<f8', (3,)),
                      ('allocated', '?'), ('f', '<f8'), ('given_time', '<f8'), ('last_given_time', '<f8'),
                      ('local_min', '?'), ('num_active_runs', '<i8'), ('cancel_requested', '?'), ('kill_sent', '?'),
                      ('given_back', '?'), ('last_given_back_time', '<f8')])

wrs2 = np.array([(0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf),
                 (0, False, 0., 0., 0., 0., 0, False, 0., False, -1, inf, inf, 0., False, False, False, inf)],
                dtype=[('gen_worker', '<i8'), ('returned', '?'), ('returned_time', '<f8'),
                       ('gen_time', '<f8'), ('last_gen_time', '<f8'), ('x', '<f8'),
                       ('sim_worker', '<i8'), ('allocated', '?'), ('g', '<f8'), ('given', '?'),
                       ('sim_id', '<i8'), ('given_time', '<f8'), ('last_given_time', '<f8'),
                       ('priority', '<f8'), ('cancel_requested', '?'), ('kill_sent', '?'),
                       ('given_back', '?'), ('last_given_back_time', '<f8')])

exp_x_in_setup2 = np.array([(0, 0, 2, 0., 4.17022005e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 1, 3, 0., 7.20324493e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 2, 3, 0., 1.14374817e-04, False, False, False, inf, 0., False, False, False, inf),
                            (0, 3, 3, 0., 3.02332573e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 4, 3, 0., 1.46755891e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 5, 3, 0., 9.23385948e-02, False, False, False, inf, 0., False, False, False, inf),
                            (0, 6, 3, 0., 1.86260211e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 7, 3, 0., 3.45560727e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 8, 3, 0., 3.96767474e-01, False, False, False, inf, 0., False, False, False, inf),
                            (0, 9, 3, 0., 5.38816734e-01, False, False, False, inf, 0., False, False, False, inf)],
                           dtype=[('sim_worker', '<i8'), ('sim_id', '<i8'), ('gen_worker', '<i8'), ('priority', '<f8'),
                                  ('x', '<f8'), ('allocated', '?'), ('returned', '?'), ('given', '?'),
                                  ('given_time', '<f8'), ('g', '<f8'), ('cancel_requested', '?'), ('kill_sent', '?'),
                                  ('given_back', '?'), ('last_given_back_time', '<f8')])

safe_mode = True


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


# Tests ========================================================================================
def test_hist_init_1():
    hist, _, _, _, _ = setup.hist_setup1()
    assert np.array_equal(hist.H, wrs), "Array does not match expected"
    assert hist.given_count == 0
    assert hist.index == 0
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_hist_init_1A_H0():
    hist, _, _, _, _ = setup.hist_setup1(sim_max=2, H0_in=wrs_H0)

    # Compare by column
    for field in exp_H0_H.dtype.names:
        np.array_equal(hist.H[field], exp_H0_H[field])
    # These dont work for numpy structured arrays
    # assert np.array_equiv(hist.H, exp_H0_H), "Array does not match expected"
    # assert np.array_equal(hist.H, exp_H0_H), "Array does not match expected"
    assert hist.given_count == 3
    assert hist.index == 3
    assert hist.returned_count == 3
    assert hist.given_back_count == 0
    assert len(hist.H) == 5


def test_hist_init_2():
    hist, _, _, _, _ = setup.hist_setup2()
    assert np.array_equal(hist.H, wrs2), "Array does not match expected"
    assert hist.given_count == 0
    assert hist.index == 0
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_grow_H():
    hist, _, _, _, _ = setup.hist_setup1(3)
    new_rows = 7
    hist.grow_H(k=new_rows)
    assert np.array_equal(hist.H, wrs), "Array does not match expected"
    assert hist.given_count == 0
    assert hist.index == 0
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_trim_H():
    hist, _, _, _, _ = setup.hist_setup1(13)
    hist.index = 10
    H = hist.trim_H()
    assert np.array_equal(H, wrs), "Array does not match expected"
    assert hist.given_count == 0
    assert hist.index == 10
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_update_history_x_in_Oempty():
    hist, sim_specs, gen_specs, _, _ = setup.hist_setup2()
    H_o = np.zeros(0, dtype=gen_specs['out'])
    gen_worker = 1
    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    assert np.array_equal(hist.H, wrs2), "H Array does not match expected"
    assert hist.given_count == 0
    assert hist.index == 0
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_update_history_x_in():
    hist, _, gen_specs, _, _ = setup.hist_setup2(7)

    np.random.seed(1)
    single_rand = gen_specs['gen_f']()  # np.random.uniform()

    # Check seeded correctly going in
    assert isclose(single_rand, 0.417022004702574), "Random numbers not correct before function"

    size = 1
    gen_worker = 2
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = single_rand

    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    assert isclose(single_rand, hist.H['x'][0])
    assert hist.given_count == 0
    assert hist.index == 1
    assert hist.returned_count == 0
    assert hist.given_back_count == 0

    size = 6
    gen_worker = 3
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = gen_specs['gen_f'](size=size)

    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    # Compare by column
    exp_x = exp_x_in_setup2[:size+1]
    for field in exp_x.dtype.names:
        np.allclose(hist.H[field], exp_x[field])

    assert hist.given_count == 0
    assert hist.index == 7
    assert hist.returned_count == 0
    assert hist.given_back_count == 0

    # Force H to grow when add points
    size = 3
    gen_worker = 3
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = gen_specs['gen_f'](size=size)

    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    # Compare by column
    exp_x = exp_x_in_setup2
    for field in exp_x.dtype.names:
        np.allclose(hist.H[field], exp_x[field])

    assert hist.given_count == 0
    assert hist.index == 10
    assert hist.returned_count == 0

    # Test libE errors when a protected field appears in output from a gen_worker
    H_o = np.zeros(size, dtype=gen_specs['out'] + [('given', bool)])
    try:
        hist.update_history_x_in(gen_worker, H_o, safe_mode)
    except AssertionError:
        assert 1, "Failed like it should have"
    else:
        assert 0, "Didn't fail like it should have"

    # Test libE errors when a protected field appears in output from a gen_worker
    H_o = np.zeros(size, dtype=gen_specs['out'] + [('given', bool)])
    try:
        hist.update_history_x_in(gen_worker, H_o, safe_mode)
    except AssertionError:
        assert 1, "Failed like it should have"
    else:
        assert 0, "Didn't fail like it should have"


def test_update_history_x_in_sim_ids():
    hist, _, gen_specs, _, _ = setup.hist_setup2A_genout_sim_ids(7)

    np.random.seed(1)
    single_rand = gen_specs['gen_f']()  # np.random.uniform()

    # Check seeded correctly going in
    assert isclose(single_rand, 0.417022004702574), "Random numbers not correct before function"

    size = 1
    gen_worker = 2
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = single_rand
    H_o['sim_id'] = 0

    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    assert isclose(single_rand, hist.H['x'][0])
    assert hist.given_count == 0
    assert hist.index == 1
    assert hist.returned_count == 0
    assert hist.given_back_count == 0

    size = 6
    gen_worker = 3
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = gen_specs['gen_f'](size=size)
    H_o['sim_id'] = range(1, 7)
    hist.update_history_x_in(gen_worker, H_o, safe_mode)

    # Compare by column
    exp_x = exp_x_in_setup2[:size+1]
    for field in exp_x.dtype.names:
        np.allclose(hist.H[field], exp_x[field])

    assert hist.given_count == 0
    assert hist.index == 7
    assert hist.returned_count == 0
    assert hist.given_back_count == 0

    # Force H to grow when add points
    size = 3
    gen_worker = 3
    H_o = np.zeros(size, dtype=gen_specs['out'])
    H_o['x'] = gen_specs['gen_f'](size=size)
    H_o['sim_id'] = range(7, 10)

    hist.update_history_x_in(gen_worker, H_o, safe_mode)
    # Compare by column
    exp_x = exp_x_in_setup2
    for field in exp_x.dtype.names:
        np.allclose(hist.H[field], exp_x[field])

    assert hist.given_count == 0
    assert hist.index == 10
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


# Note - Ideally have more setup here (so hist.index reflects generated points)
def test_update_history_x_out():
    hist, _, _, _, _ = setup.hist_setup1()

    # First update a single point
    hist.update_history_x_out(q_inds=0, sim_worker=2)

    # Check updated values for point and counts
    assert hist.H['given'][0]
    assert hist.H['sim_worker'][0] == 2
    assert hist.given_count == 1

    # Check some unchanged values for point and counts
    assert hist.index == 0
    assert hist.returned_count == 0
    hist.H['returned'][0] = False
    hist.H['allocated'][0] = False
    hist.H['f'][0] == 0.0
    hist.H['sim_id'][0] == -1

    # Check the rest of H is unaffected
    assert np.array_equal(hist.H[1:10], wrs[1:10]), "H Array slice does not match expected"

    # Update two further consecutive points
    my_qinds = np.arange(1, 3)
    hist.update_history_x_out(q_inds=my_qinds, sim_worker=3)

    # Check updated values for point and counts
    assert np.all(hist.H['given'][0:3])  # Include previous point
    assert np.all(hist.H['sim_worker'][my_qinds] == 3)
    assert hist.given_count == 3

    # Update three further non-consecutive points
    my_qinds = np.array([4, 7, 9])
    hist.update_history_x_out(q_inds=my_qinds, sim_worker=4)

    # Try to avoid tautological testing - compare columns
    assert np.array_equal(hist.H['given'], np.array([True, True, True, False, True, False, False, True, False, True]))
    assert np.array_equal(hist.H['sim_worker'], np.array([2, 3, 3, 0, 4, 0, 0, 4, 0, 4]))
    assert np.all(~hist.H['returned'])  # Should still be unaffected.

    # Check counts
    assert hist.given_count == 6
    assert hist.index == 0  # In real case this would be ahead.....
    assert hist.returned_count == 0
    assert hist.given_back_count == 0


def test_update_history_f():
    hist, sim_specs, _, _, _ = setup.hist_setup2()
    exp_vals = [0.0] * 10

    # First update a single point
    size = 1
    sim_ids = [0]  # First row to be filled
    calc_out = np.zeros(size, dtype=sim_specs['out'])
    a = np.arange(9) - 4
    calc_out['g'] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_vals[0] = calc_out['g'][0]
    D_recv = {'calc_out': calc_out,
              'persis_info': {},
              'libE_info': {'H_rows': sim_ids},
              'calc_status': WORKER_DONE,
              'calc_type': 2}

    hist.update_history_f(D_recv, safe_mode)
    assert isclose(exp_vals[0], hist.H['g'][0])
    assert np.all(hist.H['returned'][0:1])
    assert np.all(~hist.H['returned'][1:10])  # Check the rest
    assert hist.returned_count == 1
    assert hist.given_count == 0  # In real case this would be ahead.....
    assert hist.index == 0  # In real case this would be ahead....

    # Update two further consecutive points
    size = 2
    sim_ids = [1, 2]  # First row to be filled
    calc_out = np.zeros(size, dtype=sim_specs['out'])
    a = np.arange(9) - 3
    calc_out['g'][0] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_vals[1] = calc_out['g'][0]
    a = np.arange(9) - 2
    calc_out['g'][1] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_vals[2] = calc_out['g'][1]
    D_recv = {'calc_out': calc_out,
              'persis_info': {},
              'libE_info': {'H_rows': sim_ids},
              'calc_status': WORKER_DONE,
              'calc_type': 2}

    hist.update_history_f(D_recv, safe_mode)
    assert np.allclose(exp_vals, hist.H['g'])
    assert np.all(hist.H['returned'][0:3])
    assert np.all(~hist.H['returned'][3:10])  # Check the rest
    assert hist.returned_count == 3
    assert hist.given_count == 0  # In real case this would be ahead.....
    assert hist.index == 0  # In real case this would be ahead....


def test_update_history_f_vec():
    hist, sim_specs, _, _, _ = setup.hist_setup1()
    exp_fs = [0.0] * 10
    exp_fvecs = [[0.0, 0.0, 0.0]] * 10

    # First update a single point
    size = 1
    sim_ids = [0]  # First row to be filled
    calc_out = np.zeros(size, dtype=sim_specs['out'])
    a = np.array([[1, 2, 3], [-1, 1, 4]])
    calc_out['f'] = sim_specs['sim_f'](a)  # np.linalg.norm
    calc_out['fvec'] = sim_specs['sim_f'](a, axis=0)  # np.linalg.norm
    exp_fs[0] = calc_out['f'][0]
    exp_fvecs[0] = calc_out['fvec'][0]
    D_recv = {'calc_out': calc_out,
              'persis_info': {},
              'libE_info': {'H_rows': sim_ids},
              'calc_status': WORKER_DONE,
              'calc_type': 2}

    hist.update_history_f(D_recv, safe_mode)

    assert isclose(exp_fs[0], hist.H['f'][0])
    assert np.allclose(exp_fvecs[0], hist.H['fvec'][0])
    assert np.all(hist.H['returned'][0:1])
    assert np.all(~hist.H['returned'][1:10])  # Check the rest
    assert hist.returned_count == 1
    assert hist.given_count == 0  # In real case this would be ahead.....
    assert hist.index == 0  # In real case this would be ahead....

    # Update two further consecutive points
    size = 2
    sim_ids = [1, 2]  # First row to be filled
    calc_out = np.zeros(size, dtype=sim_specs['out'])

    a = np.array([[1, 3, 4], [-1, 2, 4]])
    calc_out['f'][0] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_fs[1] = calc_out['f'][0]
    calc_out['fvec'][0] = sim_specs['sim_f'](a, axis=0)  # np.linalg.norm
    exp_fvecs[1] = calc_out['fvec'][0]

    a = np.array([[2, 4, 4], [-1, 3, 4]])
    calc_out['f'][1] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_fs[2] = calc_out['f'][1]
    calc_out['fvec'][1] = sim_specs['sim_f'](a, axis=0)  # np.linalg.norm
    exp_fvecs[2] = calc_out['fvec'][1]

    D_recv = {'calc_out': calc_out,
              'persis_info': {},
              'libE_info': {'H_rows': sim_ids},
              'calc_status': WORKER_DONE,
              'calc_type': 2}

    hist.update_history_f(D_recv, safe_mode)

    assert np.allclose(exp_fs, hist.H['f'])
    assert np.allclose(exp_fvecs, hist.H['fvec'])
    assert np.all(hist.H['returned'][0:3])
    assert np.all(~hist.H['returned'][3:10])  # Check the rest
    assert hist.returned_count == 3
    assert hist.given_count == 0  # In real case this would be ahead.....
    assert hist.index == 0  # In real case this would be ahead....

    # Update two further consecutive points but with sub_array fvec components
    size = 2
    sim_ids = [3, 4]  # First row to be filled
    calc_out = np.zeros(size, dtype=[('f', float), ('fvec', float, 2)])  # Only two values

    a = np.array([[1, 3, 4], [-1, 2, 4]])
    calc_out['f'][0] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_fs[3] = calc_out['f'][0]
    calc_out['fvec'][0][0], calc_out['fvec'][0][1], _ = sim_specs['sim_f'](a, axis=0)  # np.linalg.norm
    exp_fvecs[3] = [0.0, 0.0, 0.0]  # Point to a new array - so can fill in elements
    exp_fvecs[3][:2] = calc_out['fvec'][0]  # Change first two values

    a = np.array([[2, 4, 4], [-1, 3, 4]])
    calc_out['f'][1] = sim_specs['sim_f'](a)  # np.linalg.norm
    exp_fs[4] = calc_out['f'][1]
    calc_out['fvec'][1][0], calc_out['fvec'][1][1], _ = sim_specs['sim_f'](a, axis=0)  # np.linalg.norm
    exp_fvecs[4] = [0.0, 0.0, 0.0]  # Point to a new array - so can fill in elements
    exp_fvecs[4][:2] = calc_out['fvec'][1]  # Change first two values

    D_recv = {'calc_out': calc_out,
              'persis_info': {},
              'libE_info': {'H_rows': sim_ids},
              'calc_status': WORKER_DONE,
              'calc_type': 2}

    hist.update_history_f(D_recv, safe_mode)

    assert np.allclose(exp_fs, hist.H['f'])
    assert np.allclose(exp_fvecs, hist.H['fvec'])
    assert np.all(hist.H['returned'][0:5])
    assert np.all(~hist.H['returned'][5:10])  # Check the rest
    assert hist.returned_count == 5
    assert hist.given_count == 0  # In real case this would be ahead.....
    assert hist.index == 0  # In real case this would be ahead....


def test_repack_fields():
    if 'repack_fields' in globals():
        H0 = np.zeros(3, dtype=[('g', float), ('x', float), ('large', float, 1000000)])
        assert H0.itemsize != repack_fields(H0[['x', 'g']]).itemsize, "These should not be the same size"
        assert repack_fields(H0[['x', 'g']]).itemsize < 100, "This should not be that large"


if __name__ == "__main__":
    test_hist_init_1()
    test_hist_init_1A_H0()
    test_hist_init_2()
    test_grow_H()
    test_trim_H()
    test_update_history_x_in_Oempty()
    test_update_history_x_in()
    test_update_history_x_in_sim_ids()
    test_update_history_x_out()
    test_update_history_f()
    test_update_history_f_vec()
    test_repack_fields()



"""
@Author: Created by Caleb Ju at Argonne National Labs as Given's associate,
         Summer 2021.
@About: Based on gradient sliding:
        https://link.springer.com/article/10.1007/s10107-015-0955-5
"""
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG
from libensemble.tools.consensus_subroutines import (print_final_score, get_grad,
                                                     get_consensus_gradient, get_grad_locally)


def opt_slide(H, persis_info, gen_specs, libE_info):
    """ Gradient sliding. Coordinates with alloc to do local and distributed
        (i.e., gradient of consensus step) calculations.
    """
    # Send batches until manager sends stop tag
    tag = None
    local_gen_id = persis_info['worker_num']
    # each gen has unique internal id
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    f_i_idxs = persis_info['f_i_idxs']
    n = len(ub)

    # start with random x0
    x0 = persis_info['rand_stream'].uniform(low=lb, high=ub, size=(n,))
    x_k = x0
    post_x_k = x0

    # define variables
    M = persis_info['params']['M']
    R = persis_info['params']['R']
    nu = persis_info['params']['nu']
    eps = persis_info['params']['eps']
    D = persis_info['params']['D']
    N_const = persis_info['params']['N_const']
    lam_max = persis_info['params']['lam_max']
    df_i_eval = persis_info['params'].get('df_i_eval', None)

    L = 2*R*lam_max
    N = N_const * int(((L * D)/(nu * eps))**0.5 + 1)

    if local_gen_id == 1:
        print('[{}%]: '.format(0), flush=True, end='')
    print_final_score(x_k, f_i_idxs, gen_specs, libE_info)
    percent = 0.1

    for k in range(1, N+1):
        b_k = 2.0*L/(nu * k)
        g_k = 2.0/(k+1)
        T_k = int((N*((M*k)**2)) / (D*(L**2)) + 1)

        pre_x_k = (1.0 - g_k) * post_x_k + (g_k * x_k)

        tag, Lx_k = get_consensus_gradient(pre_x_k, gen_specs, libE_info)
        if tag in [PERSIS_STOP, STOP_TAG]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        gradg = 2*R*Lx_k

        _x_k = x_k.copy()
        tag, x_k, x2_k = PS(_x_k, gradg, b_k, T_k, f_i_idxs, gen_specs, libE_info, pre_x_k, df_i_eval)

        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        post_x_k = (1.0-g_k) * post_x_k + (g_k * x2_k)

        if k/N >= percent:
            if local_gen_id == 1:
                print('[{}%]: '.format(int(percent*100)), flush=True, end='')
            percent += 0.1
            print_final_score(post_x_k, f_i_idxs, gen_specs, libE_info)

    return None, persis_info, FINISHED_PERSISTENT_GEN_TAG


def PS(x, gradg, beta, T, f_i_idxs, gen_specs, libE_info, pre_x_k, df_i_eval):
    """ Prox-sliding procedure (see https://arxiv.org/pdf/1911.10645) with
        entropy as the distance generating function for Bregman divergence.
    """
    tag = None
    u = x
    u2 = x

    for t in range(1, T+1):
        p_t = t/2.0
        theta_t = 2.0*(t+1)/(t * (t+3))

        if df_i_eval is not None:
            gradf = get_grad_locally(u, f_i_idxs, df_i_eval)
        else:
            tag, gradf = get_grad(u, f_i_idxs, gen_specs, libE_info)

        if tag in [PERSIS_STOP, STOP_TAG]:
            return tag, None, None

        u_next = get_l2_argmin(x, u, gradf, gradg, beta, p_t)
        u = u_next
        u2 = (1-theta_t) * u2 + (theta_t * u)

    return tag, u, u2


def get_l2_argmin(x, u_prev, gradf, gradg, beta, p_t):
    u_next = (beta * x) + (beta * p_t * u_prev) - gradf - gradg
    u_next = u_next / (beta * (1.0+p_t))
    return u_next

    """
    # Projection
    # X={ x : |x| <= 2 sqrt{n} }
    n = len(x)
    u_norm = la.norm(u_next, ord=2)
    B = 2*(n**0.5)
    if u_norm > B:
        u_next *= (B/u_norm)
    return u_next
    """


""" We don't use this
def get_entropy_argmin(x, u, gradf, gradg, beta, p_t):
    n = len(x)

    const = 1.0/(1+p_t) * (np.log(x) + (1+p_t) * np.ones(n, dtype=float)) - 2/(beta * (1+p_t))*gradg

    dyn = p_t/(1+p_t) * np.log(u) - 1.0/(beta * (1+p_t)) * gradf

    u_next = np.exp(const + dyn)

    return u_next
"""



"""
This module is a persistent generation function that performs a uniform
random sample when ``libE_info['persistent']`` isn't ``True``, or performs a
single persistent persistent nlopt local optimization run.
"""

__all__ = ['uniform_or_localopt']

import numpy as np

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport

import nlopt


def uniform_or_localopt(H, persis_info, gen_specs, libE_info):
    """
    This generation function returns ``gen_specs['user']['gen_batch_size']`` uniformly
    sampled points when called in nonpersistent mode (i.e., when
    ``libE_info['persistent']`` isn't ``True``).  Otherwise, the generation
    function starts a persistent nlopt local optimization run.

    .. seealso::
        `test_uniform_sampling_then_persistent_localopt_runs.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_then_persistent_localopt_runs.py>`_ # noqa
    """
    if libE_info.get('persistent'):
        x_opt, persis_info_updates, tag_out = try_and_run_nlopt(H, gen_specs, libE_info)
        H_o = []
        return H_o, persis_info_updates, tag_out
    else:
        ub = gen_specs['user']['ub']
        lb = gen_specs['user']['lb']
        n = len(lb)
        b = gen_specs['user']['gen_batch_size']

        H_o = np.zeros(b, dtype=gen_specs['out'])
        for i in range(0, b):
            x = persis_info['rand_stream'].uniform(lb, ub, (1, n))
            H_o = add_to_Out(H_o, x, i, ub, lb)

        persis_info_updates = persis_info  # Send this back so it is overwritten.
        return H_o, persis_info_updates


def try_and_run_nlopt(H, gen_specs, libE_info):
    """
    Set up objective and runs nlopt performing communication with the manager in
    order receive function values for points of interest.
    """

    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    def nlopt_obj_fun(x, grad):

        # Check if we can do an early return
        if np.array_equiv(x, H['x']):
            if gen_specs['user']['localopt_method'] in ['LD_MMA']:
                grad[:] = H['grad']
            return float(H['f'])

        # Send back x to the manager, then receive info or stop tag
        H_o = add_to_Out(np.zeros(1, dtype=gen_specs['out']), x, 0,
                         gen_specs['user']['ub'], gen_specs['user']['lb'], local=True, active=True)
        tag, Work, calc_in = ps.send_recv(H_o)
        if tag in [STOP_TAG, PERSIS_STOP]:
            raise nlopt.forced_stop

        # Return function value (and maybe gradient)
        if gen_specs['user']['localopt_method'] in ['LD_MMA']:
            grad[:] = calc_in['grad']
        return float(calc_in['f'])

    # ---------------------------------------------------------------------

    x0 = H['x'].flatten()
    lb = gen_specs['user']['lb']
    ub = gen_specs['user']['ub']
    n = len(ub)

    opt = nlopt.opt(getattr(nlopt, gen_specs['user']['localopt_method']), n)
    opt.set_lower_bounds(lb)
    opt.set_upper_bounds(ub)

    # Care must be taken with NLopt because a too-large initial step causes
    # nlopt to move the starting point!
    dist_to_bound = min(min(ub-x0), min(x0-lb))
    init_step = dist_to_bound*gen_specs['user'].get('dist_to_bound_multiple', 1)
    opt.set_initial_step(init_step)

    opt.set_maxeval(gen_specs['user'].get('localopt_maxeval', 100*n))
    opt.set_min_objective(nlopt_obj_fun)
    opt.set_xtol_rel(gen_specs['user']['xtol_rel'])

    # Run local optimization.  Only send persis_info_updates back so new
    # information added to persis_info since this persistent instance started
    # (e.g., 'run_order'), is not overwritten
    try:
        x_opt = opt.optimize(x0)
        exit_code = opt.last_optimize_result()
        persis_info_updates = {'done': True}
        if exit_code > 0 and exit_code < 5:
            persis_info_updates['x_opt'] = x_opt
    except Exception:  # Raised when manager sent PERSIS_STOP or STOP_TAG
        x_opt = []
        persis_info_updates = {}

    return x_opt, persis_info_updates, FINISHED_PERSISTENT_GEN_TAG


def add_to_Out(H_o, x, i, ub, lb, local=False, active=False):
    """
    Builds or inserts points into the numpy structured array H_o that will be sent
    back to the manager.
    """
    H_o['x'][i] = x
    H_o['x_on_cube'][i] = (x-lb)/(ub-lb)
    H_o['dist_to_unit_bounds'][i] = np.inf
    H_o['dist_to_better_l'][i] = np.inf
    H_o['dist_to_better_s'][i] = np.inf
    H_o['ind_of_better_l'][i] = -1
    H_o['ind_of_better_s'][i] = -1
    if local:
        H_o['local_pt'] = True
    if active:
        H_o['num_active_runs'] = 1

    return H_o



from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.message_numbers import (UNSET_TAG, WORKER_KILL_ON_ERR,
                                         MAN_SIGNAL_FINISH, WORKER_DONE,
                                         TASK_FAILED, WORKER_KILL_ON_TIMEOUT)
import numpy as np
import os

__all__ = ['executor_hworld']

# Alt send values through X
returned_count = 0


def custom_polling_loop(exctr, task, timeout_sec=5.0, delay=0.3):
    import time

    calc_status = UNSET_TAG  # Sim func determines status of libensemble calc - returned to worker

    while task.runtime < timeout_sec:
        time.sleep(delay)

        # print('Probing manager at time: ', task.runtime)
        exctr.manager_poll()
        if exctr.manager_signal == 'finish':
            exctr.kill(task)
            calc_status = MAN_SIGNAL_FINISH  # Worker will pick this up and close down
            print('Task {} killed by manager on worker {}'.format(task.id, exctr.workerID))
            break

        # print('Polling task at time', task.runtime)
        task.poll()
        if task.finished:
            break
        elif task.state == 'RUNNING':
            print('Task {} still running on worker {} ....'.format(task.id, exctr.workerID))

        # Check output file for error
        # print('Checking output file for error at time:', task.runtime)
        if task.stdout_exists():
            if 'Error' in task.read_stdout():
                print("Found (deliberate) Error in output file - cancelling "
                      "task {} on worker {}".format(task.id, exctr.workerID))
                exctr.kill(task)
                calc_status = WORKER_KILL_ON_ERR
                break

    # After exiting loop
    if task.finished:
        print('Task {} done on worker {}'.format(task.id, exctr.workerID))
        # Fill in calc_status if not already
        if calc_status == UNSET_TAG:
            if task.state == 'FINISHED':  # Means finished successfully
                calc_status = WORKER_DONE
            elif task.state == 'FAILED':
                calc_status = TASK_FAILED
            # elif task.state == 'USER_KILLED':
            #     calc_status = WORKER_KILL
    else:
        # assert task.state == 'RUNNING', "task.state expected to be RUNNING. Returned: " + str(task.state)
        print("Task {} timed out - killing on worker {}".format(task.id, exctr.workerID))
        exctr.kill(task)
        if task.finished:
            print('Task {} done on worker {}'.format(task.id, exctr.workerID))
        calc_status = WORKER_KILL_ON_TIMEOUT

    return task, calc_status


def executor_hworld(H, persis_info, sim_specs, libE_info):
    """ Tests launching and polling task and exiting on task finish"""
    exctr = MPIExecutor.executor
    cores = sim_specs['user']['cores']
    use_balsam = 'balsam_test' in sim_specs['user']

    args_for_sim = 'sleep 1'
    # pref send this in X as a sim_in from calling script
    global returned_count
    returned_count += 1
    timeout = 6.0
    wait = False
    launch_shc = False
    if returned_count == 1:
        args_for_sim = 'sleep 1'  # Should finish
    elif returned_count == 2:
        args_for_sim = 'sleep 1 Error'  # Worker kill on error
    if returned_count == 3:
        wait = True
        args_for_sim = 'sleep 1'  # Should finish
        launch_shc = True
    elif returned_count == 4:
        args_for_sim = 'sleep 8'  # Worker kill on timeout
        timeout = 1.0
    elif returned_count == 5:
        args_for_sim = 'sleep 2 Fail'  # Manager kill - if signal received else completes
    elif returned_count == 6:
        args_for_sim = 'sleep 60'  # Manager kill - if signal received else completes
        timeout = 65.0

    if use_balsam:
        task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim,
                            hyperthreads=True, machinefile='notused', stdout='notused',
                            wait_on_start=True)
    else:
        task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim, hyperthreads=True)

    if wait:
        task.wait()
        if not task.finished:
            calc_status = UNSET_TAG
        if task.state == 'FINISHED':
            calc_status = WORKER_DONE
        elif task.state == 'FAILED':
            calc_status = TASK_FAILED

    else:
        if returned_count >= 2 and not use_balsam:
            calc_status = exctr.polling_loop(task, timeout=timeout, delay=0.3, poll_manager=True)
            if returned_count == 2 and task.stdout_exists() and 'Error' in task.read_stdout():
                calc_status = WORKER_KILL_ON_ERR

        else:
            task, calc_status = custom_polling_loop(exctr, task, timeout)

    if use_balsam:
        task.read_file_in_workdir('ensemble.log')
        try:
            task.read_stderr()
        except ValueError:
            pass

        task = exctr.submit(app_name='sim_hump_camel_dry_run', num_procs=cores, app_args=args_for_sim,
                            hyperthreads=True, machinefile='notused', stdout='notused',
                            wait_on_start=True, dry_run=True, stage_inout=os.getcwd())

        task.poll()
        task.wait()

    # This is temp - return something - so doing six_hump_camel_func again...
    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])
    for i, x in enumerate(H['x']):
        H_o['f'][i] = six_hump_camel_func(x)
        if launch_shc:
            # Test launching a named app.
            app_args = ' '.join(str(val) for val in list(x[:]))
            task = exctr.submit(app_name='six_hump_camel', num_procs=1, app_args=app_args)
            task.wait()
            output = np.float64(task.read_stdout())
            assert np.isclose(H_o['f'][i], output)

    # This is just for testing at calling script level - status of each task
    H_o['cstat'] = calc_status

    return H_o, persis_info, calc_status


def six_hump_camel_func(x):
    """
    Definition of the six-hump camel
    """
    x1 = x[0]
    x2 = x[1]
    term1 = (4-2.1*x1**2+(x1**4)/3) * x1**2
    term2 = x1*x2
    term3 = (-4+4*x2**2) * x2**2

    return term1 + term2 + term3



"""
A test of libEnsemble exception handling.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_calc_exception.py
   python3 test_calc_exception.py --nworkers 3 --comms local
   python3 test_calc_exception.py --nworkers 3 --comms tcp
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

from libensemble.libE import libE
from libensemble.manager import LoggedException
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()


# Define sim_func
def six_hump_camel_err(H, persis_info, sim_specs, _):
    raise Exception('Deliberate error')


sim_specs = {
    'sim_f': six_hump_camel_err,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, 2)],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': 10,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'elapsed_wallclock_time': 10}

libE_specs['abort_on_exception'] = False

# Perform the run
return_flag = 1
try:
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)
except LoggedException as e:
    print("Caught deliberate exception: {}".format(e))
    return_flag = 0

if is_manager:
    assert return_flag == 0



infile = 'libE_stats.txt'
import pandas as pd

run_stats = []
with open(infile) as f:
    # content = f.readlines()
    for line in f:
        lst = line.split()
        foundstart = False
        foundend = False
        for i, val in enumerate(lst):
            if val == 'Start:':
                startdate = lst[i+1]
                starttime = lst[i+2]
                foundstart = True
            if val == 'End:':
                enddate = lst[i+1]
                endtime = lst[i+2]
                foundend = True
            if foundstart and foundend:
                run_datetime = {'start': startdate + ' ' + starttime, 'end': enddate + ' ' + endtime}
                run_stats.append(run_datetime)
                break

df = pd.DataFrame(run_stats)
df['start'] = pd.to_datetime(df['start'])
df['end'] = pd.to_datetime(df['end'])
df = df.sort_values(by='start')

time_start = df['start'][0]
dend = df.sort_values(by='end')
time_end = dend['end'].iloc[-1]
date_series = pd.date_range(time_start, time_end, freq='1S')

counts = []
for i in range(len(date_series)):
    # Inclusive/exclusive to avoid multiple accounting - need high resolution
    count = sum((df['start'] <= date_series[i]) & (date_series[i] < df['end']))
    counts.append(count)

df_list = pd.DataFrame(date_series, columns=['datetime'])

# df_count = pd.DataFrame([counts], columns=['count']) # List goes to single row by default
df_count = pd.DataFrame({'count': counts})  # Transpose to columns like this


final = df_list.join(df_count)
# print(final)

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
final.plot(x='datetime', y='count', legend=None, linewidth=2, fontsize=12)
plt.xlabel('Time', fontsize=14)
plt.ylabel('Active calculations', fontsize=14)
# plt._show()
plt.savefig('calcs_util_v_time.png', bbox='tight', transparent=True)



#!/usr/bin/env python

"""
Unit test of timers for libensemble.
"""

import time
from libensemble.utils.timer import Timer


def test_timer():
    "Test timer."

    time_start = time.time()

    timer = Timer()

    with timer:
        time.sleep(0.5)
        e1 = timer.elapsed

    e2 = timer.elapsed
    time.sleep(0.1)
    e3 = timer.elapsed
    time_mid = time.time() - time_start

    # Use external wall-clock time for upper limit to allow for system overhead
    # (e.g. virtualization, or sharing machine with other tasks)
    # assert (e1 >= 0.5) and (e1 <= 0.6), "Check timed sleep seems correct"
    assert (e1 >= 0.5) and (e1 < time_mid), "Check timed sleep within boundaries"
    assert e2 >= e1, "Check timer order."
    assert e2 == e3, "Check elapsed time stable when timer inactive."

    s1 = timer.date_start
    s2 = timer.date_end
    assert s1[0:2] == "20", "Start year is 20xx"
    assert s2[0:2] == "20", "End year is 20xx"

    s3 = "{}".format(timer)
    assert s3 == "Time: {0:.3f} Start: {1} End: {2}".format(e3, s1, s2), \
        "Check string formatting."

    time.sleep(0.2)
    time_start = time.time()
    with timer:
        time.sleep(0.5)
        total1 = timer.total

    time_end = time.time() - time_start + time_mid

    assert total1 >= 1 and total1 <= time_end, \
        "Check cumulative timing (active)."
    assert timer.total >= 1 and timer.total <= time_end, \
        "Check cumulative timing (not active)."


if __name__ == "__main__":
    test_timer()



"""
Tests libEnsemble capability to abort persistent worker.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_6-hump_camel_active_persistent_worker_abort.py
   python3 test_6-hump_camel_active_persistent_worker_abort.py --nworkers 3 --comms local
   python3 test_6-hump_camel_active_persistent_worker_abort.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np

# Import libEnsemble requirements
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f
from libensemble.gen_funcs.uniform_or_localopt import uniform_or_localopt as gen_f
from libensemble.alloc_funcs.start_persistent_local_opt_gens import start_persistent_local_opt_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.tests.regression_tests.support import uniform_or_localopt_gen_out as gen_out

nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out += [('x', float, 2), ('x_on_cube', float, 2)]
gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['x', 'f'],
    'out': gen_out,
    'user': {
        'localopt_method': 'LN_BOBYQA',
        'xtol_rel': 1e-4,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': 2,
        'dist_to_bound_multiple': 0.5,
        'localopt_maxeval': 4,
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'out': gen_out,
    'user': {
        'batch_mode': True,
        'num_active_gens': 1,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

# Set sim_max small so persistent worker is quickly terminated
exit_criteria = {'sim_max': 10, 'elapsed_wallclock_time': 300}

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert flag == 0
    save_libE_output(H, persis_info, __file__, nworkers)



import os
import glob
import yaml
import time
import numpy as np

from libensemble.executors.executor import Executor


def update_config_file(H, sim_specs):
    """
    Parameterize the configuration file for the run_openmm.py application. Values
    such as 'experiment_directory', 'task_idx', and 'stage_idx' were produced by
    the generator function (are parsed from 'H', the History array).
    """
    here = os.getcwd()

    updates = {
        'experiment_directory': os.path.abspath('../' + H['gen_dir_loc'][0]),
        'output_path': here,
        'initial_pdb_dir': here,
        'reference_pdb_file': sim_specs['user']['reference_pdb_file'],
        'simulation_length_ns': sim_specs['user']['sim_length_ns'],
        'task_idx': int(H['task_id']),
        'stage_idx': int(H['stage_id'])
    }

    config_file = sim_specs['user']['config_file']
    with open(config_file, 'r') as f:
        config = yaml.safe_load(f)

    #  Specify an unfolded pdb file if the simulation is an "initial" one
    if H['initial']:
        config['pdb_file'] = os.path.join(here, '1FME-unfolded.pdb')
    else:
        config['pdb_file'] = None

    config.update(updates)

    with open(config_file, 'w') as f:
        yaml.dump(config, f)


def run_openmm_sim_f(H, persis_info, sim_specs, libE_info):
    """ Simulation user function for running DeepDriveMD's run_openmm.py via
    the Executor. libEnsemble's worker processes call this simulation function
    with a selection of the History array that contains values produced by the
    generator function.
    """
    calc_status = 0
    os.environ["OMP_NUM_THREADS"] = '4'
    os.environ["CUDA_VISIBLE_DEVICES"] = str(libE_info['workerID'] - 1)

    #  Specify the Executor object created in the calling script.
    exctr = Executor.executor

    # Update the config file copy with relevant values
    update_config_file(H, sim_specs)

    config_file = sim_specs['user']['config_file']
    args = '-c ' + os.path.join(os.getcwd(), config_file)

    # Submit the molecular_dynamics app that was registered with the Executor.
    #  Only one process needed since bulk work presumably done on GPU.
    task = exctr.submit(app_name='molecular_dynamics', app_args=args, wait_on_start=True,
                        num_procs=1, num_nodes=1, procs_per_node=1)

    # Periodically poll our running task, then ensure the task created the expected output.
    calc_status = exctr.polling_loop(task, timeout=sim_specs['user']['sim_kill_minutes']*60, delay=1)
    time.sleep(0.2)
    assert len(glob.glob('*.h5')), 'MD Simulation did not write final output to .h5 file.'

    # Create a local History array to be populated with simulation values to send
    #  to the Manager and generator function.
    H_o = np.zeros(1, dtype=sim_specs['out'])
    H_o['sim_cstat'] = calc_status
    H_o['sim_dir_loc'] = os.getcwd().split('/')[-1]

    return H_o, persis_info, calc_status



import yaml
import pprint
import logging
import inspect
import importlib
from libensemble.libE import libE
from libensemble.alloc_funcs import defaults as alloc_defaults
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.version import __version__
from libensemble import logger

ATTR_ERR_MSG = "\n********** Unable to load \"{}\"." + \
    " Is the function or submodule correctly named? **********\n"

NOTFOUND_ERR_MSG = "\n********** Unable to load \"{}\"." + \
    " Is the package installed or the relative path correct? **********\n"


class Persis_Info:
    """
    ``persis_info`` persistent information dictionary management class. An
    instance of this (with random streams) is created on initiation of Ensemble,
    since ``persis_info`` is populated like so for most libEnsemble test-cases anyway.
    """
    def __init__(self, nworkers):
        self.nworkers = nworkers
        self.persis_info = {}

    def add_random_streams(self, num_streams=None, seed=''):
        """
        ``Persis_Info`` wrapper for ``add_unique_random_streams``. Attempt
        to simplify call, since most are identical anyway.
        """
        if num_streams:
            nstreams = num_streams
        else:
            nstreams = self.nworkers + 1

        self.persis_info = add_unique_random_streams({}, nstreams, seed=seed)
        # can access immediately, or ignore return by just using as setter
        return self.persis_info


class Ensemble:
    """
    The vast majority of libEnsemble cases require the user to instantiate
    and populate a set of specification dictionaries inside a calling script,
    call ``parse_args()``, then call ``libE()`` while passing in each spec
    dictionary. Many calling scripts and ``libE()`` calls are often identical,
    even across widely varying use-cases. This is an alternative interface for
    parameterizing libEnsemble by interacting with a class instance, and
    potentially populating it via a yaml file.
    """
    def __init__(self):
        """ Initializes an Ensemble instance. ``parse_args() called on instantiation """
        self.nworkers, self.is_manager, self.libE_specs, _ = parse_args()
        self.persis_info = Persis_Info(self.nworkers)
        self._util_logger = logging.getLogger(__name__)
        self.logger = logger
        self.logger.set_level('INFO')
        self.sim_specs = {'sim_f': None, 'in': None, 'persis_in': None, 'out': None,
                          'user': None, 'type': 'sim'}
        self.gen_specs = {'gen_f': None, 'in': None, 'persis_in': None, 'out': None,
                          'user': None, 'type': 'gen'}
        self.alloc_specs = {'alloc_f': None, 'out': None,
                            'user': None, 'type': 'alloc'}
        self.exit_criteria = {}
        self.H0 = None
        self._filename = inspect.stack()[1].filename

    def __str__(self):
        """
        Returns pretty-printed representation of Ensemble object. Depicts libEnsemble
        version, plus representations of major specification dicts.
        """
        info = '\nlibEnsemble {}\n'.format(__version__) + 79*'*' + '\n'
        info += '\nCalling Script: ' + self._filename.split('/')[-1] + '\n'

        dicts = {'libE_specs': self.libE_specs,
                 'sim_specs': self.sim_specs,
                 'gen_specs': self.gen_specs,
                 'alloc_specs': self.alloc_specs,
                 'persis_info': self.persis_info.persis_info,
                 'exit_criteria': self.exit_criteria}

        for i in dicts:
            info += '{}:\n {} \n\n'.format(i, pprint.pformat(dicts[i]))

        info += 79*'*'
        return info

    def run(self):
        """
        Initializes libEnsemble, passes in all specification dictionaries.
        Sets Ensemble instance's output H, final persis_info state, and flag.
        Spec checking (and other error handling) occurs within ``libE()``.
        """

        self.H, self.persis_info.persis_info, self.flag = \
            libE(self.sim_specs, self.gen_specs, self.exit_criteria,
                 persis_info=self.persis_info.persis_info,
                 alloc_specs=self.alloc_specs,
                 libE_specs=self.libE_specs,
                 H0=self.H0)

        return self.H, self.persis_info.persis_info, self.flag

    def _get_func(self, loaded, type):
        """ Extracts user function specified in loaded yaml dict """
        func_path_split = loaded[type + '_specs']['function'].rsplit('.', 1)
        try:
            func_name = func_path_split[-1]
            return getattr(importlib.import_module(func_path_split[0]), func_name)
        except AttributeError:
            self._util_logger.manager_warning(ATTR_ERR_MSG.format(func_name))
            raise
        except ModuleNotFoundError:
            self._util_logger.manager_warning(NOTFOUND_ERR_MSG.format(func_name))
            raise

    @staticmethod
    def _get_inputs(loaded, type):
        """ Extracts input parameters from loaded yaml dict """
        return [i for i in loaded[type + '_specs'].get('inputs', [])]

    @staticmethod
    def _get_persis_inputs(loaded, type):
        """ Extracts persis input parameters from loaded yaml dict """
        return [i for i in loaded[type + '_specs'].get('persistent_inputs', [])]

    @staticmethod
    def _get_outputs(loaded, type):
        """ Extracts output parameters from loaded yaml dict """
        outputs = loaded[type + '_specs'].get('outputs')
        if not outputs:
            return []
        fields = [i for i in outputs]
        field_params = [i for i in outputs.values()]
        results = []
        for i in range(len(fields)):
            field_type = field_params[i]['type']
            # If not a builtin datatype, *probably* a numpy datatype, e.g. "U70" for strings
            built_in_type = __builtins__.get(field_type, field_type)
            try:
                if field_params[i]['size'] == 1:
                    size = (1,)  # formatting how size=1 is typically preferred
                else:
                    size = field_params[i]['size']
                results.append((fields[i], built_in_type, size))
            except KeyError:
                results.append((fields[i], built_in_type))
        return results

    @staticmethod
    def _get_user(loaded, type):
        """ Extracts user parameters from loaded yaml dict """
        return loaded[type + '_specs'].get('user', {})

    def from_yaml(self, file):
        """ Populates libEnsemble spec dictionaries from yaml file """
        with open(file, 'r') as f:
            loaded = yaml.full_load(f)

        # Functions that extract specific values from the yaml input file
        key_funcs = {'sim_f': self._get_func,
                     'gen_f': self._get_func,
                     'alloc_f': self._get_func,
                     'in': self._get_inputs,
                     'persis_in': self._get_persis_inputs,
                     'out': self._get_outputs,
                     'user': self._get_user}

        for spec in [self.sim_specs, self.gen_specs, self.alloc_specs]:
            for key in spec:
                if key == 'type':  # should be last key. Nothing more to do
                    spec.pop('type')  # currently not a valid input
                    break
                # Lookup matching extractor, and set value to extractor's output
                try:
                    spec[key] = key_funcs[key](loaded, spec['type'])
                except KeyError as e:  # if no alloc_specs, want defaults
                    if 'alloc_specs' in e.args:
                        self.alloc_specs = alloc_defaults.alloc_specs
                        continue
                    else:
                        raise

        # exit_criteria has been included in libE_specs for space (good idea?!)
        self.exit_criteria = loaded['libE_specs']['exit_criteria']
        loaded['libE_specs'].pop('exit_criteria')

        self.libE_specs.update(loaded['libE_specs'])

    def save_output(self, file):
        """ Class wrapper for save_libE_output """
        save_libE_output(self.H, self.persis_info, file, self.nworkers)



import time
import numpy as np
import numpy.lib.recfunctions
from mpi4py import MPI

import libensemble.manager as man
import libensemble.tests.unit_tests.setup as setup

libE_specs = {'mpi_comm': MPI.COMM_WORLD}


def test_term_test_1():
    # termination_test should be True when we want to stop

    # sh - Should separate the tests
    # Test 1
    hist, sim_specs, gen_specs, exit_criteria, al = setup.hist_setup1()
    mgr = man.Manager(hist, libE_specs, al, sim_specs, gen_specs, exit_criteria)
    assert not mgr.term_test()


def test_term_test_2():
    # Test 2 - these could also be sep - with a setup or fixture....
    # Shouldn't terminate
    hist, sim_specs, gen_specs, exit_criteria, al = setup.hist_setup2()
    mgr = man.Manager(hist, libE_specs, al, sim_specs, gen_specs, exit_criteria)
    assert not mgr.term_test()
    #
    # Terminate because we've found a good 'g' value
    hist.H['g'][0] = -1
    hist.H['returned'][0] = True
    hist.index = 1
    hist.given_count = 1
    assert mgr.term_test()
    #
    # Terminate because everything has been given.
    hist.H['given'] = np.ones
    hist.given_count = len(hist.H)
    assert mgr.term_test()


def test_term_test_3():
    # Test 3.
    # Terminate because enough time has passed
    H0 = np.zeros(3, dtype=[('g', float)] + [('x', float), ('priority', float)])
    hist, sim_specs, gen_specs, exit_criteria, al = setup.hist_setup2(H0_in=H0)
    mgr = man.Manager(hist, libE_specs, al, sim_specs, gen_specs, exit_criteria)
    hist.index = 4
    hist.H['given_time'][0] = time.time()
    time.sleep(0.5)
    hist.given_count = 4
    assert mgr.term_test()


if __name__ == "__main__":
    test_term_test_1()
    test_term_test_2()
    test_term_test_3()



"""
libensemble utility class -- keeps a stack of directory locations.
"""

import os
import shutil


class LocationStack:
    """Keep a stack of directory locations."""

    def __init__(self):
        """Initialize the location dictionary and directory stack."""
        self.dirs = {}
        self.stack = []

    def copy_or_symlink(self, destdir, copy_files=[], symlink_files=[], ignore_FileExists=False):
        """ Inspired by https://stackoverflow.com/a/9793699.
        Determine paths, basenames, and conditions for copying/symlinking
        """
        for file_path in copy_files:
            file_path = os.path.expanduser(os.path.expandvars(file_path))
            src_base = os.path.basename(file_path)
            dest_path = os.path.join(destdir, src_base)
            try:
                if os.path.isdir(file_path):
                    shutil.copytree(file_path, dest_path)
                else:
                    shutil.copy(file_path, dest_path)
            except FileExistsError:
                if ignore_FileExists:
                    continue
                else:  # Indicates problem with unique sim_dirs
                    raise

        for file_path in symlink_files:
            src_path = os.path.abspath(os.path.expanduser(os.path.expandvars(file_path)))
            dest_path = os.path.join(destdir, os.path.basename(file_path))
            try:
                os.symlink(src_path, dest_path)
            except FileExistsError:
                if ignore_FileExists:
                    continue
                else:  # Indicates problem with unique sim_dirs
                    raise

    def register_loc(self, key, dirname, prefix=None, copy_files=[], symlink_files=[], ignore_FileExists=False):
        """Register a new location in the dictionary.

        Parameters
        ----------

        key:
            The key used to identify the new location.

        dirname: string:
            Directory name

        prefix: string:
            Prefix to be used with the dirname.  If prefix is not None,
            only the base part of the dirname is used.

        copy_files: list:
            Copy these files to the destination directory.

        symlink_files: list:
            Symlink these files to the destination directory.
        """
        if prefix is not None:
            prefix = os.path.expanduser(prefix)
            dirname = os.path.join(prefix, os.path.basename(dirname))

        if dirname and not os.path.isdir(dirname):
            os.makedirs(dirname, exist_ok=True)  # Prevent race-condition when no sim_dirs

        self.dirs[key] = dirname
        if len(copy_files) or len(symlink_files):
            self.copy_or_symlink(dirname, copy_files, symlink_files, ignore_FileExists)

        return dirname

    def push_loc(self, key):
        """Push a location from the dictionary."""
        self.push(self.dirs.get(key))

    def clean_locs(self):
        """Remove all directories listed in the dictionary."""
        for dirname in self.dirs.values():
            if dirname is not None and os.path.isdir(dirname):
                shutil.rmtree(dirname)

    def push(self, dirname):
        """Push the current location and change directories (if not None)."""
        if dirname is not None:
            self.stack.append(os.getcwd())
            os.chdir(dirname)
        else:
            self.stack.append(None)

    def pop(self):
        """Pop the current directory and change back."""
        dirname = self.stack.pop()
        if dirname is not None:
            os.chdir(dirname)

    class Saved:
        """Context object for use with a with statement"""
        def __init__(self, ls, dirname):
            self.ls = ls
            self.dirname = dirname

        def __enter__(self):
            self.ls.push(self.dirname)
            return self.ls

        def __exit__(self, etype, value, traceback):
            self.ls.pop()

    def loc(self, key):
        """Return a with context for pushing a location key"""
        return LocationStack.Saved(self, self.dirs.get(key))

    def dir(self, dirname):
        """Return a with context for pushing a directory"""
        return LocationStack.Saved(self, dirname)



"""
Runs libEnsemble doing uniform sampling and then evaluates those points with
varying amount of resources.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_uniform_sampling_with_variable_resources.py
   python3 test_uniform_sampling_with_variable_resources.py --nworkers 3 --comms local

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi, local
# TESTSUITE_NPROCS: 2 4

import sys
import numpy as np
import pkg_resources

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_with_variable_resources as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample_with_variable_resources as gen_f
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['sim_dirs_make'] = True
libE_specs['ensemble_dir_path'] = './ensemble_diff_nodes_w' + str(nworkers)

if libE_specs['comms'] == 'tcp':
    sys.exit("This test only runs with MPI or local -- aborting...")

# Get paths for applications to run
hello_world_app = pkg_resources.resource_filename('libensemble.sim_funcs', 'helloworld.py')
six_hump_camel_app = pkg_resources.resource_filename('libensemble.sim_funcs', 'six_hump_camel.py')

# Sim can run either helloworld or six_hump_camel
exctr = MPIExecutor()
exctr.register_app(full_path=hello_world_app, app_name='helloworld')
exctr.register_app(full_path=six_hump_camel_app, app_name='six_hump_camel')

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {'app': 'helloworld'},  # helloworld or six_hump_camel
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [
        ('priority', float),
        ('resource_sets', int),  # Set in gen func, resourced by alloc func.
        ('x', float, n),
        ('x_on_cube', float, n),
    ],
    'user': {
        'initial_batch_size': 5,
        'max_resource_sets': nworkers,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': give_sim_work_first,
    'out': [],
    'user': {
        'batch_mode': False,
        'give_all_with_same_priority': True,
        'num_active_gens': 1,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': 40, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(
    sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs, alloc_specs=alloc_specs
)

if is_manager:
    assert flag == 0
    save_libE_output(H, persis_info, __file__, nworkers)



import matplotlib
matplotlib.use('Agg')

from balsam.core import models
from matplotlib import pyplot as plt

fig, ax = plt.subplots()

plt.title('Balsam Utilization: Running Tasks v Date/Time')
plt.xlabel('Time of Day (H:M)')
plt.ylabel('Num. Tasks Running')

times, utilization = models.utilization_report()
plt.step(times, utilization, where='post')

import matplotlib.dates as mdates
myFmt = mdates.DateFormatter('%H:%M')
ax.xaxis.set_major_formatter(myFmt)
fig.autofmt_xdate()

plt.savefig('balsam_util_v_time.png')



from mpi4py import MPI

import numpy as np
import libensemble.manager as man
import libensemble.tests.unit_tests.setup as setup
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from libensemble.message_numbers import EVAL_SIM_TAG, EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, AllocException
from libensemble.tools import add_unique_random_streams
from libensemble.history import History
from libensemble.resources.scheduler import ResourceScheduler
from libensemble.resources.resources import Resources

al = {'alloc_f': give_sim_work_first, 'out': []}
libE_specs = {'mpi_comm': MPI.COMM_WORLD}
H0 = []

W = np.array([(1, 0, 0, 0, False), (2, 0, 0, 0, False),
              (3, 0, 0, 0, False), (4, 0, 0, 0, False)],
             dtype=[('worker_id', '<i8'), ('active', '<i8'), ('persis_state', '<i8'),
                    ('active_recv', '<i8'), ('zero_resource_worker', '?')])

H = np.array([(False, np.inf, False, 1, 0., 1., False, 1.6e09, 0, 1, False, [-0.49, 0.88], np.inf, 0., 0, False, False, [0., 0.], 1.6e09, np.inf),
              (False, np.inf, False, 1, 0., 1., False, 1.6e09, 0, 1, False, [-2.99, -0.79], np.inf, 0., 1, False, False, [0., 0.], 1.6e09, np.inf),
              (False, np.inf, False, 1, 0., 1., False, 1.6e09, 0, 1, False, [-2.11, -1.63], np.inf, 0., 2, False, False, [0., 0.], 1.6e09, np.inf),
              (False, np.inf, False, 1, 0., 1., False, 1.6e09, 0, 1, False, [-1.88, -0.61], np.inf, 0., 3, False, False, [0., 0.], 1.6e09, np.inf),
              (False, np.inf, False, 1, 0., 1., False, 1.6e09, 0, 1, False, [-0.61, 0.15], np.inf, 0., 4, False, False, [0., 0.], 1.6e09, np.inf)],
             dtype=[('given', '?'), ('last_given_back_time', '<f8'), ('given_back', '?'), ('gen_worker', '<i8'), ('returned_time', '<f8'),
                    ('priority', '<f8'), ('kill_sent', '?'), ('gen_time', '<f8'), ('sim_worker', '<i8'), ('resource_sets', '<i8'), ('returned', '?'),
                    ('x', '<f8', (2,)), ('last_given_time', '<f8'), ('f', '<f8'), ('sim_id', '<i8'), ('cancel_requested', '?'), ('allocated', '?'),
                    ('x_on_cube', '<f8', (2,)), ('last_gen_time', '<f8'), ('given_time', '<f8')])


def initialize_resources():
    Resources.init_resources({'comms': 'local', 'nworkers': 4, 'num_resource_sets': 4})
    Resources.resources.set_resource_manager(4)


def clear_resources():
    Resources.resources = None


def test_decide_work_and_resources():

    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_1()
    hist = History(al, sim_specs, gen_specs, exit_criteria, H0)

    mgr = man.Manager(hist, libE_specs, al, sim_specs, gen_specs, exit_criteria)
    W = mgr.W

    libE_info = {'sim_max_given': False, 'any_idle_workers': True, 'use_resource_sets': False}

    # Don't give out work when all workers are active
    W['active'] = 1
    Work, persis_info = al['alloc_f'](W, hist.H, sim_specs, gen_specs, al, {}, libE_info)
    assert len(Work) == 0


def test_als_init_normal():

    als = AllocSupport(W, True)
    assert als.manage_resources, \
        "AllocSupport instance should be managing resources for sim_work and gen_work."

    assert not als.resources, \
        "AllocSupport instance shouldn\'t be assigned a Resources object if Resources " + \
        "has not been initialized."


def test_als_init_withresources():
    initialize_resources()

    als = AllocSupport(W, True)
    assert isinstance(als.resources, Resources), \
        "AllocSupport instance didn\'t get assigned a Resources object."
    assert isinstance(als.sched, ResourceScheduler), \
        "AllocSupport instance didn\'t get assigned a ResourceScheduler object."

    clear_resources()


def test_als_assign_resources():

    als = AllocSupport(W, True)
    assert not als.assign_resources(4), \
        "AllocSupport instance shouldn\'t assign resources if not assigned a Scheduler"

    initialize_resources()
    als = AllocSupport(W, True)
    assert len(als.assign_resources(4)), \
        "AllocSupport didn\'t return a resource set team."

    clear_resources()


def test_als_worker_ids():
    als = AllocSupport(W, True)
    assert als.avail_worker_ids() == [1, 2, 3, 4], \
        "avail_worker_ids() didn\'t return expected available worker list."

    W_ps = W.copy()
    W_ps['persis_state'] = np.array([2, 0, 0, 0])
    als = AllocSupport(W_ps, True)
    assert als.avail_worker_ids(persistent=2) == [1], \
        "avail_worker_ids() didn\'t return expected persistent worker list."

    W_ar = W.copy()
    W_ar['active_recv'] = np.array([1, 0, 0, 0])
    W_ar['persis_state'] = np.array([2, 0, 0, 0])
    als = AllocSupport(W_ar, True)
    assert als.avail_worker_ids(persistent=2, active_recv=True) == [1], \
        "avail_worker_ids() didn\'t return expected persistent worker list."

    flag = 1
    try:
        als.avail_worker_ids(active_recv=True)
    except AllocException:
        flag = 0
    assert flag == 0, "AllocSupport didn't error on invalid options for avail_worker_ids()"

    W_ar = W.copy()
    W_ar['active_recv'] = np.array([1, 0, 0, 0])
    W_ar['persis_state'] = np.array([2, 0, 0, 0])
    als = AllocSupport(W_ar, True)
    assert als.avail_worker_ids(persistent=EVAL_GEN_TAG, active_recv=True) == [1], \
        "avail_worker_ids() didn\'t return expected persistent worker list."

    W_zrw = W.copy()
    W_zrw['zero_resource_worker'] = np.array([1, 0, 0, 0])
    als = AllocSupport(W_zrw, True)
    assert als.avail_worker_ids(zero_resource_workers=True) == [1], \
        "avail_worker_ids() didn\'t return expected zero resource worker list."


def test_als_evaluate_gens():
    W_gens = W.copy()
    W_gens['active'] = np.array([2, 0, 2, 0])
    als = AllocSupport(W_gens, True)
    assert als.count_gens() == 2, \
        "count_gens() didn't return correct number of active generators"

    assert als.test_any_gen(), \
        "test_any_gen() didn't return True on a generator worker being active."

    W_gens['persis_state'] = np.array([2, 0, 0, 0])

    assert als.count_persis_gens() == 1, \
        "count_persis_gens() didn't return correct number of active persistent generators"


def test_als_sim_work():
    persis_info = add_unique_random_streams({}, 5)
    als = AllocSupport(W, True)
    Work = {}
    Work[1] = als.sim_work(1, H, ['x'], np.array([0, 1, 2, 3, 4]), persis_info[1])
    assert Work[1]['H_fields'] == ['x'], \
        "H_fields were not assigned to Work dict correctly."

    assert Work[1]['persis_info'] == persis_info[1], \
        "persis_info was not assigned to Work dict correctly."

    assert Work[1]['tag'] == EVAL_SIM_TAG, \
        "sim_work didn't assign tag to EVAL_SIM_TAG"

    assert not Work[1]['libE_info']['rset_team'], \
        "rset_team should not be defined if Resources hasn\'t been initialized?"

    assert all([i in Work[1]['libE_info']['H_rows'] for i in np.array([0, 1, 2, 3, 4])]), \
        "H_rows weren't assigned to libE_info correctly."

    W_ps = W.copy()
    W_ps['persis_state'] = np.array([1, 0, 0, 0])
    als = AllocSupport(W_ps, True)
    Work = {}
    Work[1] = als.sim_work(1, H, ['x'], np.array([0, 1, 2, 3, 4]), persis_info[1], persistent=True)

    assert not len(Work[1]['libE_info']['rset_team']), \
        "Resource set should be empty for persistent workers."

    initialize_resources()
    als = AllocSupport(W, True)
    Work = {}
    Work[1] = als.sim_work(1, H, ['x'], np.array([0, 1, 2, 3, 4]), persis_info[1])

    assert len(Work[1]['libE_info']['rset_team']), \
        "Resource set should be assigned in libE_info"

    clear_resources()


def test_als_gen_work():
    persis_info = add_unique_random_streams({}, 5)
    als = AllocSupport(W, True)
    Work = {}
    Work[1] = als.gen_work(1, ['sim_id'], range(0, 5), persis_info[1])
    assert Work[1]['H_fields'] == ['sim_id'], \
        "H_fields were not assigned to Work dict correctly."

    assert Work[1]['persis_info'] == persis_info[1], \
        "persis_info was not assigned to Work dict correctly."

    assert Work[1]['tag'] == EVAL_GEN_TAG, \
        "gen_work didn't assign tag to EVAL_GEN_TAG"

    assert not Work[1]['libE_info']['rset_team'], \
        "rset_team should not be defined if Resources hasn\'t been initialized?"

    assert all([i in Work[1]['libE_info']['H_rows'] for i in range(0, 5)]), \
        "H_rows weren't assigned to libE_info correctly."

    W_ps = W.copy()
    W_ps['persis_state'] = np.array([2, 0, 0, 0])
    als = AllocSupport(W_ps, True)
    Work = {}
    Work[1] = als.gen_work(1, ['sim_id'], range(0, 5), persis_info[1], persistent=True)

    assert not len(Work[1]['libE_info']['rset_team']), \
        "Resource set should be empty for persistent workers."

    initialize_resources()
    persis_info['gen_resources'] = 1
    als = AllocSupport(W, True, persis_info=persis_info)
    Work = {}
    Work[1] = als.gen_work(1, ['sim_id'], range(0, 5), persis_info[1])

    assert len(Work[1]['libE_info']['rset_team']), \
        "Resource set should be assigned in libE_info"

    clear_resources()


def test_als_all_given():
    H_some_gvn = H.copy()
    H_some_gvn['given'] = np.array([True, False, False, True, True])
    als = AllocSupport(W, True)

    myfilter1 = np.array([False, True, False, True, True])
    myfilter2 = np.array([True, False, False, False, True])

    assert not als.all_given(H_some_gvn), \
        "all_given() should\'ve returned False on incomplete sim results."

    assert als.all_given(H_some_gvn, low_bound=3), \
        "all_given() should\'ve returned True with adjusted lower bound."

    assert not als.all_given(H_some_gvn, pt_filter=myfilter1), \
        "all_given() should\'ve returned False based on boolean filter."

    assert als.all_given(H_some_gvn, pt_filter=myfilter1, low_bound=3), \
        "all_given() should\'ve returned True with boolean filter and adjusted lower bound."

    assert als.all_given(H_some_gvn, pt_filter=myfilter2), \
        "all_given() should\'ve returned True based on boolean filter."

    # Now cancel a point
    H_some_gvn['cancel_requested'] = np.array([False, True, False, False, False])

    assert not als.all_given(H_some_gvn), \
        "all_given(H_some_gvn) should\'ve returned False on incomplete sim results."

    assert als.all_given(H_some_gvn, pt_filter=myfilter1), \
        "all_given(H_some_gvn) should\'ve returned True based on boolean filter and cancelled."

    # Now cancel more points
    H_some_gvn['cancel_requested'] = np.array([False, True, True, False, False])

    assert als.all_given(H_some_gvn), \
        "all_given(H_some_gvn) should\'ve returned True as cancelled point ignored."

    # Now cancel a different point
    H_some_gvn['cancel_requested'] = np.array([False, False, True, False, False])

    assert not als.all_given(H_some_gvn), \
        "all_given(H_some_gvn) should\'ve returned False on incomplete sim results."

    assert als.all_given(H_some_gvn, low_bound=2), \
        "all_given(H_some_gvn) should\'ve returned True with cancelled and adjusted lower bound."


def test_als_all_returned():
    H_some_rtn = H.copy()
    H_some_rtn['returned'] = np.array([True, False, False, True, True])
    als = AllocSupport(W, True)

    myfilter1 = np.array([False, True, False, True, True])
    myfilter2 = np.array([True, False, False, False, True])

    assert not als.all_returned(H_some_rtn), \
        "all_returned() should\'ve returned False on incomplete sim results."

    assert als.all_returned(H_some_rtn, low_bound=3), \
        "all_returned() should\'ve returned True with adjusted lower bound."

    assert not als.all_returned(H_some_rtn, pt_filter=myfilter1), \
        "all_returned() should\'ve returned False based on boolean filter."

    assert als.all_returned(H_some_rtn, pt_filter=myfilter1, low_bound=3), \
        "all_returned() should\'ve returned True with boolean filter and adjusted lower bound."

    assert als.all_returned(H_some_rtn, pt_filter=myfilter2), \
        "all_returned() should\'ve returned True based on boolean filter."

    # Now cancel a point
    H_some_rtn['cancel_requested'] = np.array([False, True, False, False, False])

    assert not als.all_returned(H_some_rtn), \
        "all_returned() should\'ve returned False on incomplete sim results."

    assert als.all_returned(H_some_rtn, pt_filter=myfilter1), \
        "all_returned() should\'ve returned True based on boolean filter and cancelled."

    # Now cancel more points
    H_some_rtn['cancel_requested'] = np.array([False, True, True, False, False])

    assert als.all_returned(H_some_rtn), \
        "all_returned() should\'ve returned True as cancelled point ignored."

    # Now cancel a different point
    H_some_rtn['cancel_requested'] = np.array([False, False, True, False, False])

    assert not als.all_returned(H_some_rtn), \
        "all_returned() should\'ve returned False on incomplete sim results."

    assert als.all_returned(H_some_rtn, low_bound=2), \
        "all_returned() should\'ve returned True with cancelled and adjusted lower bound."

    # But if the cancelled point is already given, we still expect it back.
    H_some_rtn['given'] = np.array([False, False, True, False, False])

    assert not als.all_returned(H_some_rtn, low_bound=2), \
        "all_returned() should\'ve returned False with given cancelled and adjusted lower bound."


def test_als_all_given_back():
    H_some_gvnbk = H.copy()
    H_some_gvnbk['given_back'] = np.array([True, False, False, True, True])
    als = AllocSupport(W, True)

    myfilter1 = np.array([False, True, False, True, True])
    myfilter2 = np.array([True, False, False, False, True])

    assert not als.all_given_back(H_some_gvnbk), \
        "all_given_back() should\'ve returned False on incomplete sim results."

    assert als.all_given_back(H_some_gvnbk, low_bound=3), \
        "all_given_back() should\'ve returned True with adjusted lower bound."

    assert not als.all_given_back(H_some_gvnbk, pt_filter=myfilter1), \
        "all_given_back() should\'ve returned False based on boolean filter."

    assert als.all_given_back(H_some_gvnbk, pt_filter=myfilter1, low_bound=3), \
        "all_given_back() should\'ve returned True with boolean filter and adjusted lower bound."

    assert als.all_given_back(H_some_gvnbk, pt_filter=myfilter2), \
        "all_given_back() should\'ve returned True based on boolean filter."

    # Now cancel a point
    H_some_gvnbk['cancel_requested'] = np.array([False, True, False, False, False])

    assert not als.all_given_back(H_some_gvnbk), \
        "all_given_back() should\'ve returned False on incomplete sim results."

    assert als.all_given_back(H_some_gvnbk, pt_filter=myfilter1), \
        "all_given_back() should\'ve returned True based on boolean filter and cancelled."

    # Now cancel more points
    H_some_gvnbk['cancel_requested'] = np.array([False, True, True, False, False])

    assert als.all_given_back(H_some_gvnbk), \
        "all_given_back() should\'ve returned True as cancelled point ignored."

    # Now cancel a different point
    H_some_gvnbk['cancel_requested'] = np.array([False, False, True, False, False])

    assert not als.all_given_back(H_some_gvnbk), \
        "all_given_back() should\'ve returned False on incomplete sim results."

    assert als.all_given_back(H_some_gvnbk, low_bound=2), \
        "all_given_back() should\'ve returned True with cancelled and adjusted lower bound."

    # But if the cancelled point is already given, we still expect it back.
    H_some_gvnbk['given'] = np.array([False, False, True, False, False])

    assert not als.all_given_back(H_some_gvnbk, low_bound=2), \
        "all_given_back() should\'ve returned False with given cancelled and adjusted lower bound."


def test_als_points_by_priority():
    H_prio = H.copy()
    H_prio['priority'] = np.array([1, 2, 1, 2, 1])
    H_no_prio = H[[i for i in list(H.dtype.names) if i != 'priority']]
    eval_pts = ~H_prio['given'] & ~H_prio['cancel_requested']  # should be same for no_prio

    als = AllocSupport(W, True)
    assert all(i in als.points_by_priority(H_prio, eval_pts, batch=True) for i in np.array([1, 3])), \
        "points_by_priority() should\'ve returned a batch of higher-priority points."

    assert als.points_by_priority(H_prio, eval_pts) == 1, \
        "points_by_priority() should\'ve returned a higher-priority index."

    als = AllocSupport(W, H_no_prio)

    assert als.points_by_priority(H_no_prio, eval_pts) == 0, \
        "points_by_priority() should\'ve simply returned the next point to evaluate."


if __name__ == "__main__":
    test_decide_work_and_resources()

    test_als_init_normal()
    test_als_init_withresources()
    test_als_assign_resources()
    test_als_worker_ids()
    test_als_evaluate_gens()
    test_als_sim_work()
    test_als_gen_work()
    test_als_all_given()
    test_als_all_returned()
    test_als_all_given_back()
    test_als_points_by_priority()



import numpy as np


def sim_find_sine(H, persis_info, sim_specs, _):
    # underscore for internal/testing arguments

    # Create an output array of a single zero
    out = np.zeros(1, dtype=sim_specs['out'])

    # Set the zero to the sine of the input value stored in H
    out['y'] = np.sin(H['x'])

    # Send back our output and persis_info
    return out, persis_info



# Sim_func
__all__ = ['likelihood_calculator']

import numpy as np


def likelihood_calculator(H, persis_info, sim_specs, _):
    """
    Evaluates likelihood
    """
    H_o = np.zeros(len(H['x']), dtype=sim_specs['out'])
    for i, x in enumerate(H['x']):
        H_o['like'][i] = six_hump_camel_func(x)

    return H_o, persis_info


def six_hump_camel_func(x):
    """
    Definition of the six-hump camel
    """
    x1 = x[0]
    x2 = x[1]
    term1 = (4-2.1*x1**2+(x1**4)/3) * x1**2
    term2 = x1*x2
    term3 = (-4+4*x2**2) * x2**2

    return term1 + term2 + term3



import numpy as np


def EvaluateFunction(x, component):
    """
    Evaluates the chained Rosenbrock function
    """
    n = len(x)

    i = component
    assert 0 <= i <= n

    if i == 0:
        x_1 = x[i]
        f_i = 0.5*x_1**2 - x_1
    elif i == n:
        x_n = x[-1]
        f_i = 0.5*x_n**2
    else:
        x_1 = x[i-1]
        x_2 = x[i]
        f_i = 0.5*(x_2-x_1)**2

    return f_i


def EvaluateJacobian(x, component=np.nan):
    """
    Evaluates the chained Rosenbrock Jacobian
    """
    n = len(x)
    df = np.zeros(n, dtype=float)

    i = component
    assert 0 <= i <= n

    if i == 0:
        x_1 = x[i]
        df[0] = x_1 - 1
    elif i == n:
        x_n = x[-1]
        df[-1] = x_n
    else:
        x_1 = x[i-1]
        x_2 = x[i]

        df[i-1] = x_1-x_2
        df[i] = x_2-x_1

    return df


def nesterov_quadratic_eval(H, persis_info, sim_specs, _):

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        obj_component = H['obj_component'][i]  # which f_i

        if H[i]['get_grad']:
            H_o['gradf_i'][i] = EvaluateJacobian(x, obj_component)
        else:
            H_o['f_i'][i] = EvaluateFunction(x, obj_component)

    return H_o, persis_info



#!/usr/bin/env python
import os             # for adding to path
import time
from mpi4py import MPI

import balsam.launcher.dag as dag


def poll_until_state(job, state, timeout_sec=60.0, delay=2.0):
    start = time.time()
    while time.time() - start < timeout_sec:
        time.sleep(delay)
        job.refresh_from_db()
        if job.state == state:
            return True
    raise RuntimeError("Task %s failed to reach state %s in %.1f seconds" % (job.cute_id, state, timeout_sec))


myrank = MPI.COMM_WORLD.Get_rank()
steps = 3
sleep_time = 3  # + myrank

# Create output dir
script_name = os.path.splitext(os.path.basename(__file__))[0]
sim_input_dir = 'simdir_' + script_name.split("test_", 1).pop()
dir_path = os.path.dirname(os.path.realpath(__file__))
sim_path = os.path.join(dir_path, sim_input_dir)

if myrank == 0:
    if not os.path.isdir(sim_path):
        try:
            os.mkdir(sim_path)
        except Exception as e:
            print(e)
            raise("Cannot make simulation directory %s" % sim_path)
MPI.COMM_WORLD.Barrier()  # Ensure output dir created

print("Host job rank is %d Output dir is %s" % (myrank, sim_input_dir))

start = time.time()
for sim_id in range(steps):
    jobname = 'outfile_t1_' + 'for_sim_id_' + str(sim_id) + '_ranks_' + str(myrank) + '.txt'

    current_job = dag.add_job(name=jobname,
                              workflow="libe_workflow",
                              application="helloworld",
                              application_args=str(sleep_time),
                              num_nodes=1,
                              procs_per_node=8,
                              stage_out_url="local:" + sim_path,
                              stage_out_files=jobname + ".out")

    success = poll_until_state(current_job, 'JOB_FINISHED')  # OR job killed
    if success:
        print("Completed job: %s rank=%d time=%f" % (jobname, myrank, time.time()-start))
    else:
        print("Task not completed: %s rank=%d time=%f Status" % (jobname, myrank, time.time()-start), current_job.state)

end = time.time()
print("Done: rank=%d  time=%f" % (myrank, end-start))



import numpy as np
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def give_sim_work_first(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function gives (in order) entries in ``H`` to idle workers
    to evaluate in the simulation function. The fields in ``sim_specs['in']``
    are given. If all entries in `H` have been given a be evaluated, a worker
    is told to call the generator function, provided this wouldn't result in
    more than ``alloc_specs['user']['num_active_gen']`` active generators. Also allows
    for a 'batch_mode'.

    When there are multiple objective components, this allocation function
    does not evaluate further components for some point in the following
    scenarios:

    alloc_specs['user']['stop_on_NaNs']: True --- after a NaN has been found in returned in some
        objective component
    alloc_specs['user']['stop_partial_fvec_eval']: True --- after the value returned from
        combine_component_func is larger than a known upper bound on the objective.

    .. seealso::
        `test_uniform_sampling_one_residual_at_a_time.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_one_residual_at_a_time.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}
    gen_count = support.count_gens()

    if gen_specs['user'].get('single_component_at_a_time'):
        assert alloc_specs['user']['batch_mode'], ("Must be in batch mode when using 'single_component_at_a_time'")
    if len(H) != persis_info['H_len']:
        # Something new is in the history.
        persis_info['need_to_give'].update(H['sim_id'][persis_info['H_len']:].tolist())
        persis_info['H_len'] = len(H)
        persis_info['pt_ids'] = set(np.unique(H['pt_id']))
        for pt_id in persis_info['pt_ids']:
            persis_info['inds_of_pt_ids'][pt_id] = H['pt_id'] == pt_id

    idle_workers = support.avail_worker_ids()

    while len(idle_workers):

        pt_ids_to_pause = set()

        # Find indices of H that are not yet given out to be evaluated
        if len(persis_info['need_to_give']):
            # If 'stop_on_NaN' is true and any f_i is a NaN, then pause
            # evaluations of other f_i, corresponding to the same pt_id
            if alloc_specs['user'].get('stop_on_NaNs'):
                pt_ids_to_pause.update(H['pt_id'][np.isnan(H['f_i'])])

            # If 'stop_partial_fvec_eval' is true, pause entries in H if a
            # partial combine_component_func evaluation is # worse than the
            # best, known, complete evaluation (and the point is not a
            # local_pt).
            if alloc_specs['user'].get('stop_partial_fvec_eval'):
                pt_ids = set(persis_info['pt_ids']) - persis_info['has_nan'] - persis_info['complete']
                pt_ids = np.array(list(pt_ids))
                partial_fvals = np.zeros(len(pt_ids))

                # Mark 'complete' and 'has_nan' pt_ids, compute complete and partial fvals
                for j, pt_id in enumerate(pt_ids):

                    a1 = persis_info['inds_of_pt_ids'][pt_id]
                    if np.any(np.isnan(H['f_i'][a1])):
                        persis_info['has_nan'].add(pt_id)
                        continue

                    if 'local_pt' in H.dtype.names and H['local_pt'][a1][0]:
                        persis_info['local_pt_ids'].add(pt_id)

                    if np.all(H['returned'][a1]):
                        persis_info['complete'].add(pt_id)
                        values = gen_specs['user']['combine_component_func'](H['f_i'][a1])
                        persis_info['best_complete_val'] = min(persis_info['best_complete_val'], values)
                    else:
                        # Ensure combine_component_func calculates partial fevals correctly
                        # with H['f_i'] = 0 for non-returned point
                        partial_fvals[j] = gen_specs['user']['combine_component_func'](H['f_i'][a1])

                if len(persis_info['complete']) and len(pt_ids) > 1:

                    worse_flag = np.zeros(len(pt_ids), dtype=bool)
                    for j, pt_id in enumerate((pt_ids)):
                        if (not np.isnan(partial_fvals[j])) and \
                           (pt_id not in persis_info['local_pt_ids']) and \
                           (pt_id not in persis_info['complete']) and \
                           (partial_fvals[j] > persis_info['best_complete_val']):
                            worse_flag[j] = True

                    # Pause incomplete evaluations with worse_flag==True
                    pt_ids_to_pause.update(pt_ids[worse_flag])

            if not pt_ids_to_pause.issubset(persis_info['already_paused']):
                persis_info['already_paused'].update(pt_ids_to_pause)
                sim_ids_to_remove = np.in1d(H['pt_id'], list(pt_ids_to_pause))
                H['paused'][sim_ids_to_remove] = True

                persis_info['need_to_give'] = persis_info['need_to_give'].difference(np.where(sim_ids_to_remove)[0])

            if len(persis_info['need_to_give']) != 0:
                next_row = persis_info['need_to_give'].pop()
                i = idle_workers[0]
                try:
                    Work[i] = support.sim_work(i, H, sim_specs['in'], [next_row], [])
                except InsufficientFreeResources:
                    persis_info['need_to_give'].add(next_row)
                    break
                idle_workers = idle_workers[1:]

        elif gen_count < alloc_specs['user'].get('num_active_gens', gen_count + 1):
            lw = persis_info['last_worker']

            last_size = persis_info.get('last_size')
            if len(H):
                # Don't give gen instances in batch mode if points are unfinished
                if (alloc_specs['user'].get('batch_mode')
                        and not all(np.logical_or(H['returned'][last_size:], H['paused'][last_size:]))):
                    break
                # Don't call APOSMM if there are runs going but none need advancing
                if len(persis_info[lw]['run_order']):
                    runs_needing_to_advance = np.zeros(len(persis_info[lw]['run_order']), dtype=bool)
                    for run, inds in enumerate(persis_info[lw]['run_order'].values()):
                        runs_needing_to_advance[run] = np.all(H['returned'][inds])

                    if not np.any(runs_needing_to_advance):
                        break

            # Give gen work
            i = idle_workers[0]
            try:
                Work[i] = support.gen_work(i, gen_specs['in'], range(len(H)), persis_info[lw])
            except InsufficientFreeResources:
                break
            idle_workers = idle_workers[1:]
            gen_count += 1
            persis_info['total_gen_calls'] += 1
            persis_info['last_worker'] = i
            persis_info['last_size'] = len(H)

        elif gen_count >= alloc_specs['user'].get('num_active_gens', gen_count + 1):
            idle_workers = []

    return Work, persis_info
